# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,cyclic-import
"""Definition of factories to load classes from the various plugin groups."""
from inspect import isclass
from typing import TYPE_CHECKING, Any, Callable, Optional, Tuple, Type, Union

from importlib_metadata import EntryPoint

from aiida.common.exceptions import InvalidEntryPointTypeError

__all__ = (
    'BaseFactory', 'CalculationFactory', 'CalcJobImporterFactory', 'DataFactory', 'DbImporterFactory', 'GroupFactory',
    'OrbitalFactory', 'ParserFactory', 'SchedulerFactory', 'TransportFactory', 'WorkflowFactory'
)

if TYPE_CHECKING:
    from aiida.engine import CalcJob, CalcJobImporter, WorkChain
    from aiida.orm import Data, Group
    from aiida.parsers import Parser
    from aiida.schedulers import Scheduler
    from aiida.tools.data.orbital import Orbital
    from aiida.tools.dbimporters import DbImporter
    from aiida.transports import Transport


def raise_invalid_type_error(entry_point_name: str, entry_point_group: str, valid_classes: Tuple[Any, ...]) -> None:
    """Raise an `InvalidEntryPointTypeError` with formatted message.

    :param entry_point_name: name of the entry point
    :param entry_point_group: name of the entry point group
    :param valid_classes: tuple of valid classes for the given entry point group
    :raises aiida.common.InvalidEntryPointTypeError: always
    """
    template = 'entry point `{}` registered in group `{}` is invalid because its type is not one of: {}'
    args = (entry_point_name, entry_point_group, ', '.join([e.__name__ for e in valid_classes]))
    raise InvalidEntryPointTypeError(template.format(*args))


def BaseFactory(group: str, name: str, load: bool = True) -> Union[EntryPoint, Any]:
    """Return the plugin class registered under a given entry point group and name.

    :param group: entry point group
    :param name: entry point name
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: the plugin class
    :raises aiida.common.MissingEntryPointError: entry point was not registered
    :raises aiida.common.MultipleEntryPointError: entry point could not be uniquely resolved
    :raises aiida.common.LoadingEntryPointError: entry point could not be loaded
    """
    from .entry_point import get_entry_point, load_entry_point

    if load is True:
        return load_entry_point(group, name)

    return get_entry_point(group, name)


def CalculationFactory(entry_point_name: str, load: bool = True) -> Optional[Union[EntryPoint, 'CalcJob', Callable]]:
    """Return the `CalcJob` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.engine.processes.calcjobs.calcjob.CalcJob`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from aiida.engine import CalcJob, calcfunction, is_process_function
    from aiida.orm import CalcFunctionNode

    entry_point_group = 'aiida.calculations'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (CalcJob, calcfunction)

    if not load:
        return entry_point

    if ((isclass(entry_point) and issubclass(entry_point, CalcJob)) or
        (is_process_function(entry_point) and entry_point.node_class is CalcFunctionNode)):  # type: ignore[union-attr]
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


def CalcJobImporterFactory(entry_point_name: str, load: bool = True) -> Optional[Union[EntryPoint, 'CalcJobImporter']]:
    """Return the plugin registered under the given entry point.

    :param entry_point_name: the entry point name.
    :return: the loaded :class:`~aiida.engine.processes.calcjobs.importer.CalcJobImporter` plugin.
    :raises ``aiida.common.InvalidEntryPointTypeError``: if the type of the loaded entry point is invalid.
    """
    from aiida.engine import CalcJobImporter

    entry_point_group = 'aiida.calculations.importers'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (CalcJobImporter,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, CalcJobImporter):
        return entry_point  # type: ignore[return-value]

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


def DataFactory(entry_point_name: str, load: bool = True) -> Optional[Union[EntryPoint, 'Data']]:
    """Return the `Data` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.orm.nodes.data.data.Data`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from aiida.orm import Data

    entry_point_group = 'aiida.data'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Data,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Data):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


def DbImporterFactory(entry_point_name: str, load: bool = True) -> Optional[Union[EntryPoint, 'DbImporter']]:
    """Return the `DbImporter` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.tools.dbimporters.baseclasses.DbImporter`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from aiida.tools.dbimporters import DbImporter

    entry_point_group = 'aiida.tools.dbimporters'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (DbImporter,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, DbImporter):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


def GroupFactory(entry_point_name: str, load: bool = True) -> Optional[Union[EntryPoint, 'Group']]:
    """Return the `Group` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.orm.groups.Group`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from aiida.orm import Group

    entry_point_group = 'aiida.groups'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Group,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Group):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


def OrbitalFactory(entry_point_name: str, load: bool = True) -> Optional[Union[EntryPoint, 'Orbital']]:
    """Return the `Orbital` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.tools.data.orbital.orbital.Orbital`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from aiida.tools.data.orbital import Orbital

    entry_point_group = 'aiida.tools.data.orbitals'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Orbital,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Orbital):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


def ParserFactory(entry_point_name: str, load: bool = True) -> Optional[Union[EntryPoint, 'Parser']]:
    """Return the `Parser` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.parsers.parser.Parser`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from aiida.parsers import Parser

    entry_point_group = 'aiida.parsers'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Parser,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Parser):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


def SchedulerFactory(entry_point_name: str, load: bool = True) -> Optional[Union[EntryPoint, 'Scheduler']]:
    """Return the `Scheduler` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.schedulers.scheduler.Scheduler`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from aiida.schedulers import Scheduler

    entry_point_group = 'aiida.schedulers'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Scheduler,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Scheduler):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


def TransportFactory(entry_point_name: str, load: bool = True) -> Optional[Union[EntryPoint, Type['Transport']]]:
    """Return the `Transport` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from aiida.transports import Transport

    entry_point_group = 'aiida.transports'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (Transport,)

    if not load:
        return entry_point

    if isclass(entry_point) and issubclass(entry_point, Transport):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


def WorkflowFactory(entry_point_name: str, load: bool = True) -> Optional[Union[EntryPoint, 'WorkChain', Callable]]:
    """Return the `WorkChain` sub class registered under the given entry point.

    :param entry_point_name: the entry point name.
    :param load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    :return: sub class of :py:class:`~aiida.engine.processes.workchains.workchain.WorkChain` or a `workfunction`
    :raises aiida.common.InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from aiida.engine import WorkChain, is_process_function, workfunction
    from aiida.orm import WorkFunctionNode

    entry_point_group = 'aiida.workflows'
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (WorkChain, workfunction)

    if not load:
        return entry_point

    if ((isclass(entry_point) and issubclass(entry_point, WorkChain)) or
        (is_process_function(entry_point) and entry_point.node_class is WorkFunctionNode)):  # type: ignore[union-attr]
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities related to the ORM."""

# AUTO-GENERATED

# yapf: disable
# pylint: disable=wildcard-import

from .calcjob import *
from .links import *
from .loaders import *
from .managers import *
from .node import *

__all__ = (
    'AbstractNodeMeta',
    'AttributeManager',
    'CalcJobResultManager',
    'CalculationEntityLoader',
    'CodeEntityLoader',
    'ComputerEntityLoader',
    'GroupEntityLoader',
    'LinkManager',
    'LinkPair',
    'LinkTriple',
    'NodeEntityLoader',
    'NodeLinksManager',
    'OrmEntityLoader',
    'get_loader',
    'get_query_type_from_type_string',
    'get_type_string_from_class',
    'load_code',
    'load_computer',
    'load_entity',
    'load_group',
    'load_node',
    'load_node_class',
    'validate_link',
)

# yapf: enable



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Implementation of the CalcJob process."""
import dataclasses
import io
import json
import os
import shutil
from typing import Any, Dict, Hashable, Optional, Type, Union

import plumpy.ports
import plumpy.process_states

from aiida import orm
from aiida.common import AttributeDict, exceptions
from aiida.common.datastructures import CalcInfo
from aiida.common.folders import Folder
from aiida.common.lang import classproperty, override
from aiida.common.links import LinkType

from ..exit_code import ExitCode
from ..ports import PortNamespace
from ..process import Process, ProcessState
from ..process_spec import CalcJobProcessSpec
from .importer import CalcJobImporter
from .tasks import UPLOAD_COMMAND, Waiting

__all__ = ('CalcJob',)


def validate_calc_job(inputs: Any, ctx: PortNamespace) -> Optional[str]:  # pylint: disable=too-many-return-statements
    """Validate the entire set of inputs passed to the `CalcJob` constructor.

    Reasons that will cause this validation to raise an `InputValidationError`:

     * No `Computer` has been specified, neither directly in `metadata.computer` nor indirectly through the `Code` input
     * The specified computer is not stored
     * The `Computer` specified in `metadata.computer` is not the same as that of the specified `Code`
     * No `Code` has been specified and no `remote_folder` input has been specified, i.e. this is no import run

    :return: string with error message in case the inputs are invalid
    """
    try:
        ctx.get_port('code')
        ctx.get_port('metadata.computer')
    except ValueError:
        # If the namespace no longer contains the `code` or `metadata.computer` ports we skip validation
        return None

    remote_folder = inputs.get('remote_folder', None)

    if remote_folder is not None:
        # The `remote_folder` input has been specified and so this concerns an import run, which means that neither
        # a `Code` nor a `Computer` are required. However, they are allowed to be specified but will not be explicitly
        # checked for consistency.
        return None

    code = inputs.get('code', None)
    computer_from_code = code.computer
    computer_from_metadata = inputs.get('metadata', {}).get('computer', None)

    if not computer_from_code and not computer_from_metadata:
        return 'no computer has been specified in `metadata.computer` nor via `code`.'

    if computer_from_code and not computer_from_code.is_stored:
        return f'the Computer<{computer_from_code}> is not stored'

    if computer_from_metadata and not computer_from_metadata.is_stored:
        return f'the Computer<{computer_from_metadata}> is not stored'

    if computer_from_code and computer_from_metadata and computer_from_code.uuid != computer_from_metadata.uuid:
        return (
            'Computer<{}> explicitly defined in `metadata.computer` is different from Computer<{}> which is the '
            'computer of Code<{}> defined as the `code` input.'.format(
                computer_from_metadata, computer_from_code, code
            )
        )

    try:
        resources_port = ctx.get_port('metadata.options.resources')
    except ValueError:
        return None

    # If the resources port exists but is not required, we don't need to validate it against the computer's scheduler
    if not resources_port.required:
        return None

    computer = computer_from_code or computer_from_metadata
    scheduler = computer.get_scheduler()
    try:
        resources = inputs['metadata']['options']['resources']
    except KeyError:
        return 'input `metadata.options.resources` is required but is not specified'

    scheduler.preprocess_resources(resources, computer.get_default_mpiprocs_per_machine())

    try:
        scheduler.validate_resources(**resources)
    except ValueError as exception:
        return f'input `metadata.options.resources` is not valid for the `{scheduler}` scheduler: {exception}'

    return None


def validate_stash_options(stash_options: Any, _: Any) -> Optional[str]:
    """Validate the ``stash`` options."""
    from aiida.common.datastructures import StashMode

    target_base = stash_options.get('target_base', None)
    source_list = stash_options.get('source_list', None)
    stash_mode = stash_options.get('mode', StashMode.COPY.value)

    if not isinstance(target_base, str) or not os.path.isabs(target_base):
        return f'`metadata.options.stash.target_base` should be an absolute filepath, got: {target_base}'

    if (
        not isinstance(source_list, (list, tuple)) or
        any(not isinstance(src, str) or os.path.isabs(src) for src in source_list)
    ):
        port = 'metadata.options.stash.source_list'
        return f'`{port}` should be a list or tuple of relative filepaths, got: {source_list}'

    try:
        StashMode(stash_mode)
    except ValueError:
        port = 'metadata.options.stash.mode'
        return f'`{port}` should be a member of aiida.common.datastructures.StashMode, got: {stash_mode}'

    return None


def validate_parser(parser_name: Any, _: Any) -> Optional[str]:
    """Validate the parser.

    :return: string with error message in case the inputs are invalid
    """
    from aiida.plugins import ParserFactory

    try:
        ParserFactory(parser_name)
    except exceptions.EntryPointError as exception:
        return f'invalid parser specified: {exception}'

    return None


def validate_additional_retrieve_list(additional_retrieve_list: Any, _: Any) -> Optional[str]:
    """Validate the additional retrieve list.

    :return: string with error message in case the input is invalid.
    """
    if any(not isinstance(value, str) or os.path.isabs(value) for value in additional_retrieve_list):
        return f'`additional_retrieve_list` should only contain relative filepaths but got: {additional_retrieve_list}'

    return None


class CalcJob(Process):
    """Implementation of the CalcJob process."""

    _node_class = orm.CalcJobNode
    _spec_class = CalcJobProcessSpec
    link_label_retrieved: str = 'retrieved'

    def __init__(self, *args, **kwargs) -> None:
        """Construct a CalcJob instance.

        Construct the instance only if it is a sub class of `CalcJob`, otherwise raise `InvalidOperation`.

        See documentation of :class:`aiida.engine.Process`.
        """
        if self.__class__ == CalcJob:
            raise exceptions.InvalidOperation('cannot construct or launch a base `CalcJob` class.')

        super().__init__(*args, **kwargs)

    @classmethod
    def define(cls, spec: CalcJobProcessSpec) -> None:  # type: ignore[override]
        """Define the process specification, including its inputs, outputs and known exit codes.

        Ports are added to the `metadata` input namespace (inherited from the base Process),
        and a `code` input Port, a `remote_folder` output Port and retrieved folder output Port
        are added.

        :param spec: the calculation job process spec to define.
        """
        # yapf: disable
        super().define(spec)
        spec.inputs.validator = validate_calc_job  # type: ignore[assignment]  # takes only PortNamespace not Port
        spec.input('code', valid_type=orm.Code, required=False,
            help='The `Code` to use for this job. This input is required, unless the `remote_folder` input is '
                 'specified, which means an existing job is being imported and no code will actually be run.')
        spec.input('remote_folder', valid_type=orm.RemoteData, required=False,
            help='Remote directory containing the results of an already completed calculation job without AiiDA. The '
                 'inputs should be passed to the `CalcJob` as normal but instead of launching the actual job, the '
                 'engine will recreate the input files and then proceed straight to the retrieve step where the files '
                 'of this `RemoteData` will be retrieved as if it had been actually launched through AiiDA. If a '
                 'parser is defined in the inputs, the results are parsed and attached as output nodes as usual.')
        spec.input('metadata.dry_run', valid_type=bool, default=False,
            help='When set to `True` will prepare the calculation job for submission but not actually launch it.')
        spec.input('metadata.computer', valid_type=orm.Computer, required=False,
            help='When using a "local" code, set the computer on which the calculation should be run.')
        spec.input_namespace(f'{spec.metadata_key}.{spec.options_key}', required=False)
        spec.input('metadata.options.input_filename', valid_type=str, required=False,
            help='Filename to which the input for the code that is to be run is written.')
        spec.input('metadata.options.output_filename', valid_type=str, required=False,
            help='Filename to which the content of stdout of the code that is to be run is written.')
        spec.input('metadata.options.submit_script_filename', valid_type=str, default='_aiidasubmit.sh',
            help='Filename to which the job submission script is written.')
        spec.input('metadata.options.scheduler_stdout', valid_type=str, default='_scheduler-stdout.txt',
            help='Filename to which the content of stdout of the scheduler is written.')
        spec.input('metadata.options.scheduler_stderr', valid_type=str, default='_scheduler-stderr.txt',
            help='Filename to which the content of stderr of the scheduler is written.')
        spec.input('metadata.options.resources', valid_type=dict, required=True,
            help='Set the dictionary of resources to be used by the scheduler plugin, like the number of nodes, '
                 'cpus etc. This dictionary is scheduler-plugin dependent. Look at the documentation of the '
                 'scheduler for more details.')
        spec.input('metadata.options.max_wallclock_seconds', valid_type=int, required=False,
            help='Set the wallclock in seconds asked to the scheduler')
        spec.input('metadata.options.custom_scheduler_commands', valid_type=str, default='',
            help='Set a (possibly multiline) string with the commands that the user wants to manually set for the '
                 'scheduler. The difference of this option with respect to the `prepend_text` is the position in '
                 'the scheduler submission file where such text is inserted: with this option, the string is '
                 'inserted before any non-scheduler command')
        spec.input('metadata.options.queue_name', valid_type=str, required=False,
            help='Set the name of the queue on the remote computer')
        spec.input('metadata.options.rerunnable', valid_type=bool, required=False,
            help='Determines if the calculation can be requeued / rerun.')
        spec.input('metadata.options.account', valid_type=str, required=False,
            help='Set the account to use in for the queue on the remote computer')
        spec.input('metadata.options.qos', valid_type=str, required=False,
            help='Set the quality of service to use in for the queue on the remote computer')
        spec.input('metadata.options.withmpi', valid_type=bool, default=False,
            help='Set the calculation to use mpi',)
        spec.input('metadata.options.mpirun_extra_params', valid_type=(list, tuple), default=lambda: [],
            help='Set the extra params to pass to the mpirun (or equivalent) command after the one provided in '
                 'computer.mpirun_command. Example: mpirun -np 8 extra_params[0] extra_params[1] ... exec.x',)
        spec.input('metadata.options.import_sys_environment', valid_type=bool, default=True,
            help='If set to true, the submission script will load the system environment variables',)
        spec.input('metadata.options.environment_variables', valid_type=dict, default=lambda: {},
            help='Set a dictionary of custom environment variables for this calculation',)
        spec.input('metadata.options.environment_variables_double_quotes', valid_type=bool, default=False,
            help='If set to True, use double quotes instead of single quotes to escape the environment variables '
                 'specified in ``environment_variables``.',)
        spec.input('metadata.options.priority', valid_type=str, required=False,
            help='Set the priority of the job to be queued')
        spec.input('metadata.options.max_memory_kb', valid_type=int, required=False,
            help='Set the maximum memory (in KiloBytes) to be asked to the scheduler')
        spec.input('metadata.options.prepend_text', valid_type=str, default='',
            help='Set the calculation-specific prepend text, which is going to be prepended in the scheduler-job '
                 'script, just before the code execution',)
        spec.input('metadata.options.append_text', valid_type=str, default='',
            help='Set the calculation-specific append text, which is going to be appended in the scheduler-job '
                 'script, just after the code execution',)
        spec.input('metadata.options.parser_name', valid_type=str, required=False, validator=validate_parser,
            help='Set a string for the output parser. Can be None if no output plugin is available or needed')
        spec.input('metadata.options.additional_retrieve_list', required=False,
            valid_type=(list, tuple), validator=validate_additional_retrieve_list,
            help='List of relative file paths that should be retrieved in addition to what the plugin specifies.')
        spec.input_namespace('metadata.options.stash', required=False, populate_defaults=False,
            validator=validate_stash_options,
            help='Optional directives to stash files after the calculation job has completed.')
        spec.input('metadata.options.stash.target_base', valid_type=str, required=False,
            help='The base location to where the files should be stashd. For example, for the `copy` stash mode, this '
            'should be an absolute filepath on the remote computer.')
        spec.input('metadata.options.stash.source_list', valid_type=(tuple, list), required=False,
            help='Sequence of relative filepaths representing files in the remote directory that should be stashed.')
        spec.input('metadata.options.stash.stash_mode', valid_type=str, required=False,
            help='Mode with which to perform the stashing, should be value of `aiida.common.datastructures.StashMode.')

        spec.output('remote_folder', valid_type=orm.RemoteData,
            help='Input files necessary to run the process will be stored in this folder node.')
        spec.output('remote_stash', valid_type=orm.RemoteStashData, required=False,
            help='Contents of the `stash.source_list` option are stored in this remote folder after job completion.')
        spec.output(cls.link_label_retrieved, valid_type=orm.FolderData, pass_to_parser=True,
            help='Files that are retrieved by the daemon will be stored in this node. By default the stdout and stderr '
                 'of the scheduler will be added, but one can add more by specifying them in `CalcInfo.retrieve_list`.')

        # Errors caused or returned by the scheduler
        spec.exit_code(100, 'ERROR_NO_RETRIEVED_FOLDER',
            message='The process did not have the required `retrieved` output.')
        spec.exit_code(110, 'ERROR_SCHEDULER_OUT_OF_MEMORY',
            message='The job ran out of memory.')
        spec.exit_code(120, 'ERROR_SCHEDULER_OUT_OF_WALLTIME',
            message='The job ran out of walltime.')
        # yapf: enable

    @classproperty
    def spec_options(cls):  # pylint: disable=no-self-argument
        """Return the metadata options port namespace of the process specification of this process.

        :return: options dictionary
        :rtype: dict
        """
        return cls.spec_metadata['options']  # pylint: disable=unsubscriptable-object

    @classmethod
    def get_importer(cls, entry_point_name: str = None) -> CalcJobImporter:
        """Load the `CalcJobImporter` associated with this `CalcJob` if it exists.

        By default an importer with the same entry point as the ``CalcJob`` will be loaded, however, this can be
        overridden using the ``entry_point_name`` argument.

        :param entry_point_name: optional entry point name of a ``CalcJobImporter`` to override the default.
        :return: the loaded ``CalcJobImporter``.
        :raises: if no importer class could be loaded.
        """
        from aiida.plugins import CalcJobImporterFactory
        from aiida.plugins.entry_point import get_entry_point_from_class

        if entry_point_name is None:
            _, entry_point = get_entry_point_from_class(cls.__module__, cls.__name__)
            if entry_point is not None:
                entry_point_name = entry_point.name  # type: ignore[attr-defined]

        return CalcJobImporterFactory(entry_point_name)()

    @property
    def options(self) -> AttributeDict:
        """Return the options of the metadata that were specified when this process instance was launched.

        :return: options dictionary

        """
        try:
            return self.metadata.options
        except AttributeError:
            return AttributeDict()

    @classmethod
    def get_state_classes(cls) -> Dict[Hashable, Type[plumpy.process_states.State]]:
        """A mapping of the State constants to the corresponding state class.

        Overrides the waiting state with the Calcjob specific version.
        """
        # Overwrite the waiting state
        states_map = super().get_state_classes()
        states_map[ProcessState.WAITING] = Waiting
        return states_map

    @override
    def on_terminated(self) -> None:
        """Cleanup the node by deleting the calulation job state.

        .. note:: This has to be done before calling the super because that will seal the node after we cannot change it
        """
        self.node.delete_state()
        super().on_terminated()

    @override
    def run(self) -> Union[plumpy.process_states.Stop, int, plumpy.process_states.Wait]:
        """Run the calculation job.

        This means invoking the `presubmit` and storing the temporary folder in the node's repository. Then we move the
        process in the `Wait` state, waiting for the `UPLOAD` transport task to be started.

        :returns: the `Stop` command if a dry run, int if the process has an exit status,
            `Wait` command if the calcjob is to be uploaded

        """
        if self.inputs.metadata.dry_run:  # type: ignore[union-attr]
            self._perform_dry_run()
            return plumpy.process_states.Stop(None, True)

        if 'remote_folder' in self.inputs:  # type: ignore[operator]
            exit_code = self._perform_import()
            return exit_code

        # The following conditional is required for the caching to properly work. Even if the source node has a process
        # state of `Finished` the cached process will still enter the running state. The process state will have then
        # been overridden by the engine to `Running` so we cannot check that, but if the `exit_status` is anything other
        # than `None`, it should mean this node was taken from the cache, so the process should not be rerun.
        if self.node.exit_status is not None:
            return self.node.exit_status

        # Launch the upload operation
        return plumpy.process_states.Wait(msg='Waiting to upload', data=UPLOAD_COMMAND)

    def prepare_for_submission(self, folder: Folder) -> CalcInfo:
        """Prepare the calculation for submission.

        Convert the input nodes into the corresponding input files in the format that the code will expect. In addition,
        define and return a `CalcInfo` instance, which is a simple data structure that contains  information for the
        engine, for example, on what files to copy to the remote machine, what files to retrieve once it has completed,
        specific scheduler settings and more.

        :param folder: a temporary folder on the local file system.
        :returns: the `CalcInfo` instance
        """
        raise NotImplementedError()

    def _perform_dry_run(self):
        """Perform a dry run.

        Instead of performing the normal sequence of steps, just the `presubmit` is called, which will call the method
        `prepare_for_submission` of the plugin to generate the input files based on the inputs. Then the upload action
        is called, but using a normal local transport that will copy the files to a local sandbox folder. The generated
        input script and the absolute path to the sandbox folder are stored in the `dry_run_info` attribute of the node
        of this process.
        """
        from aiida.common.folders import SubmitTestFolder
        from aiida.engine.daemon.execmanager import upload_calculation
        from aiida.transports.plugins.local import LocalTransport

        with LocalTransport() as transport:
            with SubmitTestFolder() as folder:
                calc_info = self.presubmit(folder)
                transport.chdir(folder.abspath)
                upload_calculation(self.node, transport, calc_info, folder, inputs=self.inputs, dry_run=True)
                self.node.dry_run_info = {
                    'folder': folder.abspath,
                    'script_filename': self.node.get_option('submit_script_filename')
                }

    def _perform_import(self):
        """Perform the import of an already completed calculation.

        The inputs contained a `RemoteData` under the key `remote_folder` signalling that this is not supposed to be run
        as a normal calculation job, but rather the results are already computed outside of AiiDA and merely need to be
        imported.
        """
        from aiida.common.datastructures import CalcJobState
        from aiida.common.folders import SandboxFolder
        from aiida.engine.daemon.execmanager import retrieve_calculation
        from aiida.transports.plugins.local import LocalTransport

        with LocalTransport() as transport:
            with SandboxFolder() as folder:
                with SandboxFolder() as retrieved_temporary_folder:
                    self.presubmit(folder)
                    self.node.set_remote_workdir(
                        self.inputs.remote_folder.get_remote_path()  # type: ignore[union-attr]
                    )
                    retrieve_calculation(self.node, transport, retrieved_temporary_folder.abspath)
                    self.node.set_state(CalcJobState.PARSING)
                    self.node.base.attributes.set(orm.CalcJobNode.IMMIGRATED_KEY, True)
                    return self.parse(retrieved_temporary_folder.abspath)

    def parse(self, retrieved_temporary_folder: Optional[str] = None) -> ExitCode:
        """Parse a retrieved job calculation.

        This is called once it's finished waiting for the calculation to be finished and the data has been retrieved.

        :param retrieved_temporary_folder: The path to the temporary folder

        """
        try:
            retrieved = self.node.outputs.retrieved
        except exceptions.NotExistent:
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER  # pylint: disable=no-member

        # Call the scheduler output parser
        exit_code_scheduler = self.parse_scheduler_output(retrieved)

        if exit_code_scheduler is not None and exit_code_scheduler.status > 0:
            # If an exit code is returned by the scheduler output parser, we log it and set it on the node. This will
            # allow the actual `Parser` implementation, if defined in the inputs, to inspect it and decide to keep it,
            # or override it with a more specific exit code, if applicable.
            msg = f'scheduler parser returned exit code<{exit_code_scheduler.status}>: {exit_code_scheduler.message}'
            self.logger.warning(msg)
            self.node.set_exit_status(exit_code_scheduler.status)
            self.node.set_exit_message(exit_code_scheduler.message)

        # Call the retrieved output parser
        try:
            exit_code_retrieved = self.parse_retrieved_output(retrieved_temporary_folder)
        finally:
            if retrieved_temporary_folder is not None:
                shutil.rmtree(retrieved_temporary_folder, ignore_errors=True)

        if exit_code_retrieved is not None and exit_code_retrieved.status > 0:
            msg = f'output parser returned exit code<{exit_code_retrieved.status}>: {exit_code_retrieved.message}'
            self.logger.warning(msg)

        # The final exit code is that of the scheduler, unless the output parser returned one
        exit_code: Optional[ExitCode]
        if exit_code_retrieved is not None:
            exit_code = exit_code_retrieved
        else:
            exit_code = exit_code_scheduler

        # Finally link up the outputs and we're done
        for entry in self.node.base.links.get_outgoing():
            self.out(entry.link_label, entry.node)

        return exit_code or ExitCode(0)

    def parse_scheduler_output(self, retrieved: orm.Node) -> Optional[ExitCode]:
        """Parse the output of the scheduler if that functionality has been implemented for the plugin."""
        computer = self.node.computer

        if computer is None:
            self.logger.info(
                'no computer is defined for this calculation job which suggest that it is an imported job and so '
                'scheduler output probably is not available or not in a format that can be reliably parsed, skipping..'
            )
            return None

        scheduler = computer.get_scheduler()
        filename_stderr = self.node.get_option('scheduler_stderr')
        filename_stdout = self.node.get_option('scheduler_stdout')

        detailed_job_info = self.node.get_detailed_job_info()

        if detailed_job_info is None:
            self.logger.info('could not parse scheduler output: the `detailed_job_info` attribute is missing')
        elif detailed_job_info.get('retval', 0) != 0:
            self.logger.info('could not parse scheduler output: return value of `detailed_job_info` is non-zero')
            detailed_job_info = None

        if filename_stderr is None:
            self.logger.warning('could not determine `stderr` filename because `scheduler_stderr` option was not set.')
        else:
            try:
                scheduler_stderr = retrieved.base.repository.get_object_content(filename_stderr)
            except FileNotFoundError:
                scheduler_stderr = None
                self.logger.warning(f'could not parse scheduler output: the `{filename_stderr}` file is missing')

        if filename_stdout is None:
            self.logger.warning('could not determine `stdout` filename because `scheduler_stdout` option was not set.')
        else:
            try:
                scheduler_stdout = retrieved.base.repository.get_object_content(filename_stdout)
            except FileNotFoundError:
                scheduler_stdout = None
                self.logger.warning(f'could not parse scheduler output: the `{filename_stdout}` file is missing')

        try:
            exit_code = scheduler.parse_output(detailed_job_info, scheduler_stdout, scheduler_stderr)
        except exceptions.FeatureNotAvailable:
            self.logger.info(f'`{scheduler.__class__.__name__}` does not implement scheduler output parsing')
            return None
        except Exception as exception:  # pylint: disable=broad-except
            self.logger.error(f'the `parse_output` method of the scheduler excepted: {exception}')
            return None

        if exit_code is not None and not isinstance(exit_code, ExitCode):
            args = (scheduler.__class__.__name__, type(exit_code))
            raise ValueError('`{}.parse_output` returned neither an `ExitCode` nor None, but: {}'.format(*args))

        return exit_code

    def parse_retrieved_output(self, retrieved_temporary_folder: Optional[str] = None) -> Optional[ExitCode]:
        """Parse the retrieved data by calling the parser plugin if it was defined in the inputs."""
        parser_class = self.node.get_parser_class()

        if parser_class is None:
            return None

        parser = parser_class(self.node)
        parse_kwargs = parser.get_outputs_for_parsing()

        if retrieved_temporary_folder:
            parse_kwargs['retrieved_temporary_folder'] = retrieved_temporary_folder

        exit_code = parser.parse(**parse_kwargs)

        for link_label, node in parser.outputs.items():
            try:
                self.out(link_label, node)
            except ValueError as exception:
                self.logger.error(f'invalid value {node} specified with label {link_label}: {exception}')
                exit_code = self.exit_codes.ERROR_INVALID_OUTPUT  # pylint: disable=no-member
                break

        if exit_code is not None and not isinstance(exit_code, ExitCode):
            args = (parser_class.__name__, type(exit_code))
            raise ValueError('`{}.parse` returned neither an `ExitCode` nor None, but: {}'.format(*args))

        return exit_code

    def presubmit(self, folder: Folder) -> CalcInfo:
        """Prepares the calculation folder with all inputs, ready to be copied to the cluster.

        :param folder: a SandboxFolder that can be used to write calculation input files and the scheduling script.

        :return calcinfo: the CalcInfo object containing the information needed by the daemon to handle operations.

        """
        # pylint: disable=too-many-locals,too-many-statements,too-many-branches
        from aiida.common.datastructures import CodeInfo, CodeRunMode
        from aiida.common.exceptions import InputValidationError, InvalidOperation, PluginInternalError, ValidationError
        from aiida.common.utils import validate_list_of_string_tuples
        from aiida.orm import Code, Computer, load_node
        from aiida.schedulers.datastructures import JobTemplate, JobTemplateCodeInfo

        inputs = self.node.base.links.get_incoming(link_type=LinkType.INPUT_CALC)

        if not self.inputs.metadata.dry_run and not self.node.is_stored:  # type: ignore[union-attr]
            raise InvalidOperation('calculation node is not stored.')

        computer = self.node.computer
        codes = [_ for _ in inputs.all_nodes() if isinstance(_, Code)]

        for code in codes:
            if not code.can_run_on(computer):
                raise InputValidationError(
                    'The selected code {} for calculation {} cannot run on computer {}'.format(
                        code.pk, self.node.pk, computer.label
                    )
                )

            if code.is_local() and code.get_local_executable() in folder.get_content_list():
                raise PluginInternalError(
                    f'The plugin created a file {code.get_local_executable()} that is also the executable name!'
                )

        calc_info = self.prepare_for_submission(folder)
        calc_info.uuid = str(self.node.uuid)

        # I create the job template to pass to the scheduler
        job_tmpl = JobTemplate()
        job_tmpl.submit_as_hold = False
        job_tmpl.rerunnable = self.options.get('rerunnable', False)
        job_tmpl.job_environment = {}
        # 'email', 'email_on_started', 'email_on_terminated',
        job_tmpl.job_name = f'aiida-{self.node.pk}'
        job_tmpl.sched_output_path = self.options.scheduler_stdout
        if computer is not None:
            job_tmpl.shebang = computer.get_shebang()
        if self.options.scheduler_stderr == self.options.scheduler_stdout:
            job_tmpl.sched_join_files = True
        else:
            job_tmpl.sched_error_path = self.options.scheduler_stderr
            job_tmpl.sched_join_files = False

        # Set retrieve path, add also scheduler STDOUT and STDERR
        retrieve_list = calc_info.retrieve_list or []
        if (job_tmpl.sched_output_path is not None and job_tmpl.sched_output_path not in retrieve_list):
            retrieve_list.append(job_tmpl.sched_output_path)
        if not job_tmpl.sched_join_files:
            if (job_tmpl.sched_error_path is not None and job_tmpl.sched_error_path not in retrieve_list):
                retrieve_list.append(job_tmpl.sched_error_path)
        retrieve_list.extend(self.node.get_option('additional_retrieve_list') or [])
        self.node.set_retrieve_list(retrieve_list)

        # Handle the retrieve_temporary_list
        retrieve_temporary_list = calc_info.retrieve_temporary_list or []
        self.node.set_retrieve_temporary_list(retrieve_temporary_list)

        # If the inputs contain a ``remote_folder`` input node, we are in an import scenario and can skip the rest
        if 'remote_folder' in inputs.all_link_labels():
            return

        # The remaining code is only necessary for actual runs, for example, creating the submission script
        scheduler = computer.get_scheduler()

        # the if is done so that if the method returns None, this is
        # not added. This has two advantages:
        # - it does not add too many \n\n if most of the prepend_text are empty
        # - most importantly, skips the cases in which one of the methods
        #   would return None, in which case the join method would raise
        #   an exception
        prepend_texts = [computer.get_prepend_text()] + \
            [code.get_prepend_text() for code in codes] + \
            [calc_info.prepend_text, self.node.get_option('prepend_text')]
        job_tmpl.prepend_text = '\n\n'.join(prepend_text for prepend_text in prepend_texts if prepend_text)

        append_texts = [self.node.get_option('append_text'), calc_info.append_text] + \
            [code.get_append_text() for code in codes] + \
            [computer.get_append_text()]
        job_tmpl.append_text = '\n\n'.join(append_text for append_text in append_texts if append_text)

        # Set resources, also with get_default_mpiprocs_per_machine
        resources = self.node.get_option('resources')
        scheduler.preprocess_resources(resources, computer.get_default_mpiprocs_per_machine())
        job_tmpl.job_resource = scheduler.create_job_resource(**resources)

        subst_dict = {'tot_num_mpiprocs': job_tmpl.job_resource.get_tot_num_mpiprocs()}

        for key, value in job_tmpl.job_resource.items():
            subst_dict[key] = value
        mpi_args = [arg.format(**subst_dict) for arg in computer.get_mpirun_command()]
        extra_mpirun_params = self.node.get_option('mpirun_extra_params')  # same for all codes in the same calc

        # set the codes_info
        if not isinstance(calc_info.codes_info, (list, tuple)):
            raise PluginInternalError('codes_info passed to CalcInfo must be a list of CalcInfo objects')

        tmpl_codes_info = []
        for code_info in calc_info.codes_info:

            if not isinstance(code_info, CodeInfo):
                raise PluginInternalError('Invalid codes_info, must be a list of CodeInfo objects')

            if code_info.code_uuid is None:
                raise PluginInternalError('CalcInfo should have the information of the code to be launched')
            this_code = load_node(code_info.code_uuid, sub_classes=(Code,))

            # To determine whether this code should be run with MPI enabled, we get the value that was set in the inputs
            # of the entire process, which can then be overwritten by the value from the `CodeInfo`. This allows plugins
            # to force certain codes to run without MPI, even if the user wants to run all codes with MPI whenever
            # possible. This use case is typically useful for `CalcJob`s that consist of multiple codes where one or
            # multiple codes always have to be executed without MPI.

            this_withmpi = self.node.get_option('withmpi')

            # Override the value of `withmpi` with that of the `CodeInfo` if and only if it is set
            if code_info.withmpi is not None:
                this_withmpi = code_info.withmpi

            if this_withmpi:
                prepend_cmdline_params = mpi_args + extra_mpirun_params
            else:
                prepend_cmdline_params = []

            cmdline_params = [this_code.get_execname()] + (code_info.cmdline_params or [])

            tmpl_code_info = JobTemplateCodeInfo()
            tmpl_code_info.prepend_cmdline_params = prepend_cmdline_params
            tmpl_code_info.cmdline_params = cmdline_params
            tmpl_code_info.use_double_quotes = [computer.get_use_double_quotes(), this_code.get_use_double_quotes()]
            tmpl_code_info.stdin_name = code_info.stdin_name
            tmpl_code_info.stdout_name = code_info.stdout_name
            tmpl_code_info.stderr_name = code_info.stderr_name
            tmpl_code_info.join_files = code_info.join_files

            tmpl_codes_info.append(tmpl_code_info)
        job_tmpl.codes_info = tmpl_codes_info

        # set the codes execution mode, default set to `SERIAL`
        codes_run_mode = CodeRunMode.SERIAL
        if calc_info.codes_run_mode:
            codes_run_mode = calc_info.codes_run_mode

        job_tmpl.codes_run_mode = codes_run_mode
        ########################################################################

        custom_sched_commands = self.node.get_option('custom_scheduler_commands')
        if custom_sched_commands:
            job_tmpl.custom_scheduler_commands = custom_sched_commands

        job_tmpl.import_sys_environment = self.node.get_option('import_sys_environment')

        job_tmpl.job_environment = self.node.get_option('environment_variables')
        job_tmpl.environment_variables_double_quotes = self.node.get_option('environment_variables_double_quotes')

        queue_name = self.node.get_option('queue_name')
        account = self.node.get_option('account')
        qos = self.node.get_option('qos')
        if queue_name is not None:
            job_tmpl.queue_name = queue_name
        if account is not None:
            job_tmpl.account = account
        if qos is not None:
            job_tmpl.qos = qos
        priority = self.node.get_option('priority')
        if priority is not None:
            job_tmpl.priority = priority

        job_tmpl.max_memory_kb = self.node.get_option('max_memory_kb') or computer.get_default_memory_per_machine()

        max_wallclock_seconds = self.node.get_option('max_wallclock_seconds')
        if max_wallclock_seconds is not None:
            job_tmpl.max_wallclock_seconds = max_wallclock_seconds

        submit_script_filename = self.node.get_option('submit_script_filename')
        script_content = scheduler.get_submit_script(job_tmpl)
        folder.create_file_from_filelike(io.StringIO(script_content), submit_script_filename, 'w', encoding='utf8')

        def encoder(obj):
            if dataclasses.is_dataclass(obj):
                return dataclasses.asdict(obj)
            raise TypeError(f' {obj!r} is not JSON serializable')

        subfolder = folder.get_subfolder('.aiida', create=True)
        subfolder.create_file_from_filelike(
            io.StringIO(json.dumps(job_tmpl, default=encoder)), 'job_tmpl.json', 'w', encoding='utf8'
        )
        subfolder.create_file_from_filelike(io.StringIO(json.dumps(calc_info)), 'calcinfo.json', 'w', encoding='utf8')

        if calc_info.local_copy_list is None:
            calc_info.local_copy_list = []

        if calc_info.remote_copy_list is None:
            calc_info.remote_copy_list = []

        # Some validation
        this_pk = self.node.pk if self.node.pk is not None else '[UNSTORED]'
        local_copy_list = calc_info.local_copy_list
        try:
            validate_list_of_string_tuples(local_copy_list, tuple_length=3)
        except ValidationError as exception:
            raise PluginInternalError(
                f'[presubmission of calc {this_pk}] local_copy_list format problem: {exception}'
            ) from exception

        remote_copy_list = calc_info.remote_copy_list
        try:
            validate_list_of_string_tuples(remote_copy_list, tuple_length=3)
        except ValidationError as exception:
            raise PluginInternalError(
                f'[presubmission of calc {this_pk}] remote_copy_list format problem: {exception}'
            ) from exception

        for (remote_computer_uuid, _, dest_rel_path) in remote_copy_list:
            try:
                Computer.collection.get(uuid=remote_computer_uuid)  # pylint: disable=unused-variable
            except exceptions.NotExistent as exception:
                raise PluginInternalError(
                    '[presubmission of calc {}] '
                    'The remote copy requires a computer with UUID={}'
                    'but no such computer was found in the '
                    'database'.format(this_pk, remote_computer_uuid)
                ) from exception
            if os.path.isabs(dest_rel_path):
                raise PluginInternalError(
                    '[presubmission of calc {}] '
                    'The destination path of the remote copy '
                    'is absolute! ({})'.format(this_pk, dest_rel_path)
                )

        return calc_info



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Migration from v0.11 to v0.12, used by ``verdi archive migrate`` command.

This migration is necessary after the `core.` prefix was added to entry points shipped with `aiida-core`.
"""
from ..utils import update_metadata, verify_metadata_version  # pylint: disable=no-name-in-module

MAPPING_DATA = {
    'data.array.ArrayData.': 'data.core.array.ArrayData.',
    'data.array.bands.BandsData.': 'data.core.array.bands.BandsData.',
    'data.array.kpoints.KpointsData.': 'data.core.array.kpoints.KpointsData.',
    'data.array.projection.ProjectionData.': 'data.core.array.projection.ProjectionData.',
    'data.array.trajectory.TrajectoryData.': 'data.core.array.trajectory.TrajectoryData.',
    'data.array.xy.XyData.': 'data.core.array.xy.XyData.',
    'data.base.BaseData.': 'data.core.base.BaseData.',
    'data.bool.Bool.': 'data.core.bool.Bool.',
    'data.cif.CifData.': 'data.core.cif.CifData.',
    'data.code.Code.': 'data.core.code.Code.',
    'data.dict.Dict.': 'data.core.dict.Dict.',
    'data.float.Float.': 'data.core.float.Float.',
    'data.folder.FolderData.': 'data.core.folder.FolderData.',
    'data.int.Int.': 'data.core.int.Int.',
    'data.list.List.': 'data.core.list.List.',
    'data.numeric.NumericData.': 'data.core.numeric.NumericData.',
    'data.orbital.OrbitalData.': 'data.core.orbital.OrbitalData.',
    'data.remote.RemoteData.': 'data.core.remote.RemoteData.',
    'data.remote.stash.RemoteStashData.': 'data.core.remote.stash.RemoteStashData.',
    'data.remote.stash.folder.RemoteStashFolderData.': 'data.core.remote.stash.folder.RemoteStashFolderData.',
    'data.singlefile.SinglefileData.': 'data.core.singlefile.SinglefileData.',
    'data.str.Str.': 'data.core.str.Str.',
    'data.structure.StructureData.': 'data.core.structure.StructureData.',
    'data.upf.UpfData.': 'data.core.upf.UpfData.',
}

MAPPING_SCHEDULERS = {
    'direct': 'core.direct',
    'lsf': 'core.lsf',
    'pbspro': 'core.pbspro',
    'sge': 'core.sge',
    'slurm': 'core.slurm',
    'torque': 'core.torque',
}

MAPPING_CALCULATIONS = {
    'aiida.calculations:arithmetic.add': 'aiida.calculations:core.arithmetic.add',
    'aiida.calculations:templatereplacer': 'aiida.calculations:core.templatereplacer',
}

MAPPING_PARSERS = {
    'arithmetic.add': 'core.arithmetic.add',
    'templatereplacer.doubler': 'core.templatereplacer.doubler',
}

MAPPING_WORKFLOWS = {
    'aiida.workflows:arithmetic.add_multiply': 'aiida.workflows:core.arithmetic.add_multiply',
    'aiida.workflows:arithmetic.multiply_add': 'aiida.workflows:core.arithmetic.multiply_add',
}


def migrate_v11_to_v12(metadata: dict, data: dict) -> None:
    """Migration of export files from v0.11 to v0.12."""
    # pylint: disable=too-many-branches
    old_version = '0.11'
    new_version = '0.12'

    verify_metadata_version(metadata, old_version)
    update_metadata(metadata, new_version)

    # Migrate data entry point names
    for values in data.get('export_data', {}).get('Node', {}).values():
        if 'node_type' in values and values['node_type'].startswith('data.'):
            try:
                new_node_type = MAPPING_DATA[values['node_type']]
            except KeyError:
                pass
            else:
                values['node_type'] = new_node_type

        if 'process_type' in values and values['process_type'] and values['process_type'
                                                                          ].startswith('aiida.calculations:'):
            try:
                new_process_type = MAPPING_CALCULATIONS[values['process_type']]
            except KeyError:
                pass
            else:
                values['process_type'] = new_process_type

        if 'process_type' in values and values['process_type'] and values['process_type'
                                                                          ].startswith('aiida.workflows:'):
            try:
                new_process_type = MAPPING_WORKFLOWS[values['process_type']]
            except KeyError:
                pass
            else:
                values['process_type'] = new_process_type

    for attributes in data.get('export_data', {}).get('node_attributes', {}).values():
        if 'parser_name' in attributes:
            try:
                new_parser_name = MAPPING_PARSERS[attributes['parser_name']]
            except KeyError:
                pass
            else:
                attributes['parser_name'] = new_parser_name

    # Migrate scheduler entry point names
    for values in data.get('export_data', {}).get('Computer', {}).values():
        if 'scheduler_type' in values:
            try:
                new_scheduler_type = MAPPING_SCHEDULERS[values['scheduler_type']]
            except KeyError:
                pass
            else:
                values['scheduler_type'] = new_scheduler_type



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=import-error,no-name-in-module
"""Module to manage computers for the SQLA backend."""
from sqlalchemy.dialects.postgresql import JSONB, UUID
from sqlalchemy.schema import Column
from sqlalchemy.sql.schema import Index
from sqlalchemy.types import Integer, String, Text

from aiida.common.utils import get_new_uuid
from aiida.storage.psql_dos.models.base import Base


class DbComputer(Base):
    """Database model to store data for :py:class:`aiida.orm.Computer`.

    Computers represent (and contain the information of) the physical hardware resources available.
    Nodes can be associated with computers if they are remote codes, remote folders, or processes that had run remotely.

    Computers are identified within AiiDA by their ``label`` (and thus it must be unique for each one in the database),
    whereas the ``hostname`` is the label that identifies the computer within the network from which one can access it.

    The ``scheduler_type`` column contains the information of the scheduler (and plugin)
    that the computer uses to manage jobs, whereas the ``transport_type`` the information of the transport
    (and plugin) required to copy files and communicate to and from the computer.
    The ``metadata`` contains some general settings for these communication and management protocols.
    """
    __tablename__ = 'db_dbcomputer'

    id = Column(Integer, primary_key=True)  # pylint: disable=invalid-name
    uuid = Column(UUID(as_uuid=True), default=get_new_uuid, nullable=False, unique=True)
    label = Column(String(255), nullable=False, unique=True)
    hostname = Column(String(255), default='', nullable=False)
    description = Column(Text, default='', nullable=False)
    scheduler_type = Column(String(255), default='', nullable=False)
    transport_type = Column(String(255), default='', nullable=False)
    _metadata = Column('metadata', JSONB, default=dict, nullable=False)

    __table_args__ = (
        Index(
            'ix_pat_db_dbcomputer_label',
            label,
            postgresql_using='btree',
            postgresql_ops={'label': 'varchar_pattern_ops'}
        ),
    )

    @property
    def pk(self):
        return self.id

    def __str__(self):
        return f'{self.label} ({self.hostname})'



# -*- coding: utf-8 -*-
"""Data plugin that models an archived folder on a remote computer."""
from aiida.common.datastructures import StashMode
from aiida.common.lang import type_check

from ...data import Data

__all__ = ('RemoteStashData',)


class RemoteStashData(Data):
    """Data plugin that models an archived folder on a remote computer.

    A stashed folder is essentially an instance of ``RemoteData`` that has been archived. Archiving in this context can
    simply mean copying the content of the folder to another location on the same or another filesystem as long as it is
    on the same machine. In addition, the folder may have been compressed into a single file for efficiency or even
    written to tape. The ``stash_mode`` attribute will distinguish how the folder was stashed which will allow the
    implementation to also `unstash` it and transform it back into a ``RemoteData`` such that it can be used as an input
    for new ``CalcJobs``.

    This class is a non-storable base class that merely registers the ``stash_mode`` attribute. Only its subclasses,
    that actually implement a certain stash mode, can be instantiated and therefore stored. The reason for this design
    is that because the behavior of the class can change significantly based on the mode employed to stash the files and
    implementing all these variants in the same class will lead to an unintuitive interface where certain properties or
    methods of the class will only be available or function properly based on the ``stash_mode``.
    """

    _storable = False

    def __init__(self, stash_mode: StashMode, **kwargs):
        """Construct a new instance

        :param stash_mode: the stashing mode with which the data was stashed on the remote.
        """
        super().__init__(**kwargs)
        self.stash_mode = stash_mode

    @property
    def stash_mode(self) -> StashMode:
        """Return the mode with which the data was stashed on the remote.

        :return: the stash mode.
        """
        return StashMode(self.base.attributes.get('stash_mode'))

    @stash_mode.setter
    def stash_mode(self, value: StashMode):
        """Set the mode with which the data was stashed on the remote.

        :param value: the stash mode.
        """
        type_check(value, StashMode)
        self.base.attributes.set('stash_mode', value.value)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=cyclic-import
"""Components to communicate tasks to RabbitMQ."""
import asyncio
from collections.abc import Mapping
import logging
import traceback

from kiwipy import Future, communications
import pamqp.encode
import plumpy

from aiida.common.extendeddicts import AttributeDict

__all__ = ('RemoteException', 'CommunicationTimeout', 'DeliveryFailed', 'ProcessLauncher', 'BROKER_DEFAULTS')

# The following statement enables support for RabbitMQ 3.5 because without it, connections established by `aiormq` will
# fail because the interpretation of the types of integers passed in connection parameters has changed after that
# version. Once RabbitMQ 3.5 is no longer supported (it has been EOL since October 2016) this can be removed. This
# should also allow to remove the direct dependency on `pamqp` entirely.
pamqp.encode.support_deprecated_rabbitmq()

LOGGER = logging.getLogger(__name__)

RemoteException = plumpy.RemoteException
DeliveryFailed = plumpy.DeliveryFailed
CommunicationTimeout = communications.TimeoutError  # pylint: disable=invalid-name

_LAUNCH_QUEUE = 'process.queue'
_MESSAGE_EXCHANGE = 'messages'
_TASK_EXCHANGE = 'tasks'

BROKER_DEFAULTS = AttributeDict({
    'protocol': 'amqp',
    'username': 'guest',
    'password': 'guest',
    'host': '127.0.0.1',
    'port': 5672,
    'virtual_host': '',
    'heartbeat': 600,
})


def get_rmq_url(protocol=None, username=None, password=None, host=None, port=None, virtual_host=None, **kwargs):
    """Return the URL to connect to RabbitMQ.

    .. note::

        The default of the ``host`` is set to ``127.0.0.1`` instead of ``localhost`` because on some computers localhost
        resolves first to IPv6 with address ::1 and if RMQ is not running on IPv6 one gets an annoying warning. For more
        info see: https://github.com/aiidateam/aiida-core/issues/1142

    :param protocol: the protocol to use, `amqp` or `amqps`.
    :param username: the username for authentication.
    :param password: the password for authentication.
    :param host: the hostname of the RabbitMQ server.
    :param port: the port of the RabbitMQ server.
    :param virtual_host: the virtual host to connect to.
    :param kwargs: remaining keyword arguments that will be encoded as query parameters.
    :returns: the connection URL string.
    """
    from urllib.parse import urlencode, urlunparse

    if 'heartbeat' not in kwargs:
        kwargs['heartbeat'] = BROKER_DEFAULTS.heartbeat

    scheme = protocol or BROKER_DEFAULTS.protocol
    netloc = '{username}:{password}@{host}:{port}'.format(
        username=username or BROKER_DEFAULTS.username,
        password=password or BROKER_DEFAULTS.password,
        host=host or BROKER_DEFAULTS.host,
        port=port or BROKER_DEFAULTS.port,
    )
    path = virtual_host or BROKER_DEFAULTS.virtual_host
    parameters = ''
    query = urlencode(kwargs)
    fragment = ''

    # The virtual host is optional but if it is specified it needs to start with a forward slash. If the virtual host
    # itself contains forward slashes, they need to be encoded.
    if path and not path.startswith('/'):
        path = f'/{path}'

    return urlunparse((scheme, netloc, path, parameters, query, fragment))


def get_launch_queue_name(prefix=None):
    """Return the launch queue name with an optional prefix.

    :returns: launch queue name
    """
    if prefix is not None:
        return f'{prefix}.{_LAUNCH_QUEUE}'

    return _LAUNCH_QUEUE


def get_message_exchange_name(prefix):
    """Return the message exchange name for a given prefix.

    :returns: message exchange name
    """
    return f'{prefix}.{_MESSAGE_EXCHANGE}'


def get_task_exchange_name(prefix):
    """Return the task exchange name for a given prefix.

    :returns: task exchange name
    """
    return f'{prefix}.{_TASK_EXCHANGE}'


def _store_inputs(inputs):
    """Try to store the values in the input dictionary.

    For nested dictionaries, the values are stored by recursively.
    """
    for node in inputs.values():
        try:
            node.store()
        except AttributeError:
            if isinstance(node, Mapping):
                _store_inputs(node)


class ProcessLauncher(plumpy.ProcessLauncher):
    """A sub class of `plumpy.ProcessLauncher` to launch a `Process`.

    It overrides the _continue method to make sure the node corresponding to the task can be loaded and
    that if it is already marked as terminated, it is not continued but the future is reconstructed and returned
    """

    @staticmethod
    def handle_continue_exception(node, exception, message):
        """Handle exception raised in `_continue` call.

        If the process state of the node has not yet been put to excepted, the exception was raised before the process
        instance could be reconstructed, for example when the process class could not be loaded, thereby circumventing
        the exception handling of the state machine. Raising this exception will then acknowledge the process task with
        RabbitMQ leaving an uncleaned node in the `CREATED` state for ever. Therefore we have to perform the node
        cleaning manually.

        :param exception: the exception object
        :param message: string message to use for the log message
        """
        from aiida.engine import ProcessState

        if not node.is_excepted and not node.is_sealed:
            node.logger.exception(message)
            node.set_exception(''.join(traceback.format_exception(type(exception), exception, None)).rstrip())
            node.set_process_state(ProcessState.EXCEPTED)
            node.seal()

    async def _continue(self, communicator, pid, nowait, tag=None):
        """Continue the task.

        Note that the task may already have been completed, as indicated from the corresponding the node, in which
        case it is not continued, but the corresponding future is reconstructed and returned. This scenario may
        occur when the Process was already completed by another worker that however failed to send the acknowledgment.

        :param communicator: the communicator that called this method
        :param pid: the pid of the process to continue
        :param nowait: if True don't wait for the process to finish, just return the pid, otherwise wait and
            return the results
        :param tag: the tag of the checkpoint to continue from
        """
        from aiida.common import exceptions
        from aiida.engine.exceptions import PastException
        from aiida.orm import Data, load_node
        from aiida.orm.utils import serialize

        try:
            node = load_node(pk=pid)
        except (exceptions.MultipleObjectsError, exceptions.NotExistent):
            # In this case, the process node corresponding to the process id, cannot be resolved uniquely or does not
            # exist. The latter being the most common case, where someone deleted the node, before the process was
            # properly terminated. Since the node is never coming back and so the process will never be able to continue
            # we raise `Return` instead of `TaskRejected` because the latter would cause the task to be resent and start
            # to ping-pong between RabbitMQ and the daemon workers.
            LOGGER.exception('Cannot continue process<%d>', pid)
            return False

        if node.is_terminated:

            LOGGER.info('not continuing process<%d> which is already terminated with state %s', pid, node.process_state)

            future = Future()

            if node.is_finished:
                future.set_result({
                    entry.link_label: entry.node for entry in node.base.links.get_outgoing(node_class=Data)
                })
            elif node.is_excepted:
                future.set_exception(PastException(node.exception))
            elif node.is_killed:
                future.set_exception(plumpy.KilledError())

            return future.result()

        try:
            result = await super()._continue(communicator, pid, nowait, tag)
        except ImportError as exception:
            message = 'the class of the process could not be imported.'
            self.handle_continue_exception(node, exception, message)
            raise
        except asyncio.CancelledError:  # pylint: disable=try-except-raise
            # note this is only required in python<=3.7,
            # where asyncio.CancelledError inherits from Exception
            raise
        except Exception as exception:
            message = 'failed to recreate the process instance in order to continue it.'
            self.handle_continue_exception(node, exception, message)
            raise

        # Ensure that the result is serialized such that communication thread won't have to do database operations
        try:
            serialized = serialize.serialize(result)
        except Exception:
            LOGGER.exception('failed to serialize the result for process<%d>', pid)
            raise

        return serialized



# -*- coding: utf-8 -*-
"""Data plugin that allows to easily wrap an :class:`enum.Enum` member.

Nomenclature is taken from Python documentation: https://docs.python.org/3/library/enum.html
Given the following example implementation:

.. code:: python

    from enum import Enum
    class Color(Enum):
        RED = 1
        GREEN = 2

The class ``Color`` is an enumeration (or enum). The attributes ``Color.RED`` and ``Color.GREEN`` are enumeration
members (or enum members) and are functionally constants. The enum members have names and values: the name of
``Color.RED`` is ``RED`` and the value of ``Color.RED`` is ``1``.
"""
from enum import Enum
import typing as t

from plumpy.loaders import get_object_loader

from aiida.common.lang import type_check

from .base import to_aiida_type
from .data import Data

__all__ = ('EnumData',)

EnumType = t.TypeVar('EnumType', bound=Enum)


@to_aiida_type.register(Enum)
def _(value):
    return EnumData(member=value)


class EnumData(Data):
    """Data plugin that allows to easily wrap an :class:`enum.Enum` member.

    The enum member is stored in the database by storing the value, name and the identifier (string that represents the
    class of the enumeration) in the ``KEY_NAME``, ``KEY_VALUE`` and ``KEY_IDENTIFIER`` attribute, respectively. The
    original enum member can be reconstructured from the (loaded) node through the ``get_member`` method. The enum
    itself can be retrieved from the ``get_enum`` method. Like a normal enum member, the ``EnumData`` plugin provides
    the ``name`` and ``value`` properties which return the name and value of the enum member, respectively.
    """

    KEY_NAME = 'name'
    KEY_VALUE = 'value'
    KEY_IDENTIFIER = 'identifier'

    def __init__(self, member: Enum, *args, **kwargs):
        """Construct the node for the to enum member that is to be wrapped."""
        type_check(member, Enum)
        super().__init__(*args, **kwargs)

        data = {
            self.KEY_NAME: member.name,
            self.KEY_VALUE: member.value,
            self.KEY_IDENTIFIER: get_object_loader().identify_object(member.__class__)
        }

        self.base.attributes.set_many(data)

    @property
    def name(self) -> str:
        """Return the name of the enum member."""
        return self.base.attributes.get(self.KEY_NAME)

    @property
    def value(self) -> t.Any:
        """Return the value of the enum member."""
        return self.base.attributes.get(self.KEY_VALUE)

    def get_enum(self) -> t.Type[EnumType]:
        """Return the enum class reconstructed from the serialized identifier stored in the database.

        :raises `ImportError`: if the enum class represented by the stored identifier cannot be imported.
        """
        identifier = self.base.attributes.get(self.KEY_IDENTIFIER)
        try:
            return get_object_loader().load_object(identifier)
        except ValueError as exc:
            raise ImportError(f'Could not reconstruct enum class because `{identifier}` could not be loaded.') from exc

    def get_member(self) -> EnumType:
        """Return the enum member reconstructed from the serialized data stored in the database.

        For the enum member to be successfully reconstructed, the class of course has to still be importable and its
        implementation should not have changed since the node was stored. That is to say, the value of the member when
        it was stored, should still be a valid value for the enum class now.

        :raises `ImportError`: if the enum class represented by the stored identifier cannot be imported.
        :raises `ValueError`: if the stored enum member value is no longer valid for the imported enum class.
        """
        value = self.base.attributes.get(self.KEY_VALUE)
        enum: t.Type[EnumType] = self.get_enum()

        try:
            return enum(value)
        except ValueError as exc:
            raise ValueError(
                f'The stored value `{value}` is no longer a valid value for the enum `{enum}`. The definition must '
                'have changed since storing the node.'
            ) from exc

    def __eq__(self, other: t.Any) -> bool:
        """Return whether the other object is equivalent to ourselves."""
        if isinstance(other, Enum):
            try:
                return self.get_member() == other
            except (ImportError, ValueError):
                return False
        elif isinstance(other, EnumData):
            return self.base.attributes.all == other.base.attributes.all

        return False



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Profile param type for click."""
from click.shell_completion import CompletionItem

from .strings import LabelStringType

__all__ = ('ProfileParamType',)


class ProfileParamType(LabelStringType):
    """The profile parameter type for click."""

    name = 'profile'

    def __init__(self, *args, **kwargs):
        self._cannot_exist = kwargs.pop('cannot_exist', False)
        self._load_profile = kwargs.pop('load_profile', False)  # If True, will load the profile converted from value
        super().__init__(*args, **kwargs)

    @staticmethod
    def deconvert_default(value):
        return value.name

    def convert(self, value, param, ctx):
        """Attempt to match the given value to a valid profile."""
        from aiida.common import extendeddicts
        from aiida.common.exceptions import MissingConfigurationError, ProfileConfigurationError
        from aiida.manage.configuration import Profile, get_config, load_profile

        # If the value is already of the expected return type, simply return it. This behavior is new in `click==8.0`:
        # https://click.palletsprojects.com/en/8.0.x/parameters/#implementing-custom-types
        if isinstance(value, Profile):
            return value

        value = super().convert(value, param, ctx)

        try:
            config = get_config(create=True)
            profile = config.get_profile(value)
        except (MissingConfigurationError, ProfileConfigurationError) as exception:
            if not self._cannot_exist:
                self.fail(str(exception))

            # Create a new empty profile
            profile = Profile(value, {}, validate=False)
        else:
            if self._cannot_exist:
                self.fail(str(f'the profile `{value}` already exists'))

        if self._load_profile:
            load_profile(profile.name)

        if ctx.obj is None:
            ctx.obj = extendeddicts.AttributeDict()

        ctx.obj.config = config
        ctx.obj.profile = profile

        return profile

    def shell_complete(self, ctx, param, incomplete):  # pylint: disable=unused-argument,no-self-use
        """Return possible completions based on an incomplete value

        :returns: list of tuples of valid entry points (matching incomplete) and a description
        """
        from aiida.common.exceptions import MissingConfigurationError
        from aiida.manage.configuration import get_config

        if self._cannot_exist:
            return []

        try:
            config = get_config()
        except MissingConfigurationError:
            return []

        return [CompletionItem(profile.name) for profile in config.profiles if profile.name.startswith(incomplete)]



# -*- coding: utf-8 -*-
"""Module with data plugins that represent files of completed calculations jobs that have been stashed."""

# AUTO-GENERATED

# yapf: disable
# pylint: disable=wildcard-import

from .base import *
from .folder import *

__all__ = (
    'RemoteStashData',
    'RemoteStashFolderData',
)

# yapf: enable



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Module for the custom click param type computer
"""
from click.shell_completion import CompletionItem
from click.types import StringParamType

from ...utils import decorators  # pylint: disable=no-name-in-module
from .identifier import IdentifierParamType

__all__ = ('ComputerParamType', 'ShebangParamType', 'MpirunCommandParamType')


class ComputerParamType(IdentifierParamType):
    """
    The ParamType for identifying Computer entities or its subclasses
    """

    name = 'Computer'

    @property
    def orm_class_loader(self):
        """
        Return the orm entity loader class, which should be a subclass of OrmEntityLoader. This class is supposed
        to be used to load the entity for a given identifier

        :return: the orm entity loader class for this ParamType
        """
        from aiida.orm.utils.loaders import ComputerEntityLoader
        return ComputerEntityLoader

    @decorators.with_dbenv()
    def shell_complete(self, ctx, param, incomplete):  # pylint: disable=unused-argument
        """Return possible completions based on an incomplete value.

        :returns: list of tuples of valid entry points (matching incomplete) and a description
        """
        return [CompletionItem(option) for option, in self.orm_class_loader.get_options(incomplete, project='label')]


class ShebangParamType(StringParamType):
    """
    Custom click param type for shebang line
    """
    name = 'shebangline'

    def convert(self, value, param, ctx):
        newval = super().convert(value, param, ctx)
        if newval is None:
            return None
        if not newval.startswith('#!'):
            self.fail(f'The shebang line should start with the two caracters #!, it is instead: {newval}')
        return newval

    def __repr__(self):
        return 'SHEBANGLINE'


class MpirunCommandParamType(StringParamType):
    """
    Custom click param type for mpirun-command

    .. note:: requires also a scheduler to be provided, and the scheduler
       must be called first!

    Validate that the provided 'mpirun' command only contains replacement fields
    (e.g. ``{tot_num_mpiprocs}``) that are known.

    Return a list of arguments (by using 'value.strip().split(" ") on the input string)
    """
    name = 'mpiruncommandstring'

    def __repr__(self):
        return 'MPIRUNCOMMANDSTRING'

    def convert(self, value, param, ctx):
        newval = super().convert(value, param, ctx)

        scheduler_ep = ctx.params.get('scheduler', None)
        if scheduler_ep is not None:
            try:
                job_resource_keys = scheduler_ep.load().job_resource_class.get_valid_keys()
            except ImportError:
                self.fail(f"Unable to load the '{scheduler_ep.name}' scheduler")
        else:
            self.fail(
                'Scheduler not specified for this computer! The mpirun-command must always be asked '
                'after asking for the scheduler.'
            )

        # Prepare some substitution values to check if it is all ok
        subst = {i: 'value' for i in job_resource_keys}
        subst['tot_num_mpiprocs'] = 'value'

        try:
            newval.format(**subst)
        except KeyError as exc:
            self.fail(f"In workdir there is an unknown replacement field '{exc.args[0]}'")
        except ValueError as exc:
            self.fail(f"Error in the string: '{exc}'")

        return newval



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=too-many-arguments,import-error,too-many-locals,broad-except
"""`verdi archive` command."""
from enum import Enum
import logging
from pathlib import Path
import traceback
from typing import List, Tuple
import urllib.request

import click
from click_spinner import spinner

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options
from aiida.cmdline.params.types import GroupParamType, PathOrUrl
from aiida.cmdline.utils import decorators, echo
from aiida.common.exceptions import CorruptStorage, IncompatibleStorageSchema, UnreachableStorage
from aiida.common.links import GraphTraversalRules
from aiida.common.log import AIIDA_LOGGER

EXTRAS_MODE_EXISTING = ['keep_existing', 'update_existing', 'mirror', 'none']
EXTRAS_MODE_NEW = ['import', 'none']
COMMENT_MODE = ['leave', 'newest', 'overwrite']


@verdi.group('archive')
def verdi_archive():
    """Create, inspect and import AiiDA archives."""


@verdi_archive.command('version')
@click.argument('path', nargs=1, type=click.Path(exists=True, readable=True))
def archive_version(path):
    """Print the current version of an archive's schema."""
    # note: this mirrors `cmd_storage:storage_version`
    # it is currently hardcoded to the `SqliteZipBackend`, but could be generalized in the future
    from aiida.storage.sqlite_zip.backend import SqliteZipBackend
    storage_cls = SqliteZipBackend
    profile = storage_cls.create_profile(path)
    head_version = storage_cls.version_head()
    try:
        profile_version = storage_cls.version_profile(profile)
    except (UnreachableStorage, CorruptStorage) as exc:
        echo.echo_critical(f'archive file version unreadable: {exc}')
    echo.echo(f'Latest archive schema version: {head_version!r}')
    echo.echo(f'Archive schema version of {Path(path).name!r}: {profile_version!r}')


@verdi_archive.command('info')
@click.argument('path', nargs=1, type=click.Path(exists=True, readable=True))
@click.option('--detailed', is_flag=True, help='Provides more detailed information.')
def archive_info(path, detailed):
    """Summarise the contents of an archive."""
    # note: this mirrors `cmd_storage:storage_info`
    # it is currently hardcoded to the `SqliteZipBackend`, but could be generalized in the future
    from aiida.storage.sqlite_zip.backend import SqliteZipBackend
    try:
        storage = SqliteZipBackend(SqliteZipBackend.create_profile(path))
    except (UnreachableStorage, CorruptStorage) as exc:
        echo.echo_critical(f'archive file unreadable: {exc}')
    except IncompatibleStorageSchema as exc:
        echo.echo_critical(f'archive version incompatible: {exc}')
    with spinner():
        try:
            data = storage.get_info(detailed=detailed)
        finally:
            storage.close()

    echo.echo_dictionary(data, sort_keys=False, fmt='yaml')


@verdi_archive.command('inspect', hidden=True)
@click.argument('archive', nargs=1, type=click.Path(exists=True, readable=True))
@click.option('-v', '--version', is_flag=True, help='Print the archive format version and exit.')
@click.option('-m', '--meta-data', is_flag=True, help='Print the meta data contents and exit.')
@click.option('-d', '--database', is_flag=True, help='Include information on entities in the database.')
@decorators.deprecated_command(
    'This command has been deprecated and will be removed soon. '
    'Please call `verdi archive version` or `verdi archive info` instead.\n'
)
@click.pass_context
def inspect(ctx, archive, version, meta_data, database):  # pylint: disable=unused-argument
    """Inspect contents of an archive without importing it.

    .. deprecated:: v2.0.0, use `verdi archive version` or `verdi archive info` instead.
    """
    if version:
        ctx.invoke(archive_version, path=archive)
    elif database:
        ctx.invoke(archive_info, path=archive, detailed=True)
    else:
        ctx.invoke(archive_info, path=archive, detailed=False)


@verdi_archive.command('create')
@arguments.OUTPUT_FILE(type=click.Path(exists=False))
@options.ALL()
@options.CODES()
@options.COMPUTERS()
@options.GROUPS()
@options.NODES()
@options.FORCE(help='Overwrite output file if it already exists.')
@options.graph_traversal_rules(GraphTraversalRules.EXPORT.value)
@click.option(
    '--include-logs/--exclude-logs',
    default=True,
    show_default=True,
    help='Include or exclude logs for node(s) in export.'
)
@click.option(
    '--include-comments/--exclude-comments',
    default=True,
    show_default=True,
    help='Include or exclude comments for node(s) in export. (Will also export extra users who commented).'
)
@click.option(
    '--include-authinfos/--exclude-authinfos',
    default=False,
    show_default=True,
    help='Include or exclude authentication information for computer(s) in export.'
)
@click.option('--compress', default=6, show_default=True, type=int, help='Level of compression to use (0-9).')
@click.option(
    '-b', '--batch-size', default=1000, type=int, help='Stream database rows in batches, to reduce memory usage.'
)
@click.option('--test-run', is_flag=True, help='Determine entities to export, but do not create the archive.')
@decorators.with_dbenv()
def create(
    output_file, all_entries, codes, computers, groups, nodes, force, input_calc_forward, input_work_forward,
    create_backward, return_backward, call_calc_backward, call_work_backward, include_comments, include_logs,
    include_authinfos, compress, batch_size, test_run
):
    """Create an archive from all or part of a profiles's data.

    Besides Nodes of the provenance graph, you can archive Groups, Codes, Computers, Comments and Logs.

    By default, the archive file will include not only the entities explicitly provided via the command line but also
    their provenance, according to the rules outlined in the documentation.
    You can modify some of those rules using options of this command.
    """
    # pylint: disable=too-many-branches
    from aiida.common.progress_reporter import set_progress_bar_tqdm, set_progress_reporter
    from aiida.tools.archive.abstract import get_format
    from aiida.tools.archive.create import create_archive
    from aiida.tools.archive.exceptions import ArchiveExportError

    archive_format = get_format()

    if all_entries:
        entities = None
    else:
        entities = []

        if codes:
            entities.extend(codes)

        if computers:
            entities.extend(computers)

        if groups:
            entities.extend(groups)

        if nodes:
            entities.extend(nodes)

    kwargs = {
        'input_calc_forward': input_calc_forward,
        'input_work_forward': input_work_forward,
        'create_backward': create_backward,
        'return_backward': return_backward,
        'call_calc_backward': call_calc_backward,
        'call_work_backward': call_work_backward,
        'include_authinfos': include_authinfos,
        'include_comments': include_comments,
        'include_logs': include_logs,
        'overwrite': force,
        'compression': compress,
        'batch_size': batch_size,
        'test_run': test_run
    }

    if AIIDA_LOGGER.level <= logging.REPORT:  # pylint: disable=no-member
        set_progress_bar_tqdm(leave=(AIIDA_LOGGER.level <= logging.INFO))
    else:
        set_progress_reporter(None)

    try:
        create_archive(entities, filename=output_file, archive_format=archive_format, **kwargs)
    except ArchiveExportError as exception:
        echo.echo_critical(f'failed to write the archive file. Exception: {exception}')
    else:
        echo.echo_success(f'wrote the export archive file to {output_file}')


@verdi_archive.command('migrate')
@arguments.INPUT_FILE()
@arguments.OUTPUT_FILE(required=False)
@options.FORCE(help='overwrite output file if it already exists')
@click.option('-i', '--in-place', is_flag=True, help='Migrate the archive in place, overwriting the original file.')
@click.option(
    '--version',
    type=click.STRING,
    required=False,
    metavar='VERSION',
    # Note: Adding aiida.tools.EXPORT_VERSION as a default value explicitly would result in a slow import of
    # aiida.tools and, as a consequence, aiida.orm. As long as this is the case, better determine the latest export
    # version inside the function when needed.
    help='Archive format version to migrate to (defaults to latest version).',
)
def migrate(input_file, output_file, force, in_place, version):
    """Migrate an archive to a more recent schema version."""
    from aiida.common.progress_reporter import set_progress_bar_tqdm, set_progress_reporter
    from aiida.tools.archive.abstract import get_format

    if in_place:
        if output_file:
            echo.echo_critical('output file specified together with --in-place flag')
        output_file = input_file
        force = True
    elif not output_file:
        echo.echo_critical(
            'no output file specified. Please add --in-place flag if you would like to migrate in place.'
        )

    if AIIDA_LOGGER.level <= logging.REPORT:  # pylint: disable=no-member
        set_progress_bar_tqdm(leave=(AIIDA_LOGGER.level <= logging.INFO))
    else:
        set_progress_reporter(None)

    archive_format = get_format()

    if version is None:
        version = archive_format.latest_version

    try:
        archive_format.migrate(input_file, output_file, version, force=force, compression=6)
    except Exception as error:  # pylint: disable=broad-except
        if AIIDA_LOGGER.level <= logging.DEBUG:
            raise
        echo.echo_critical(
            'failed to migrate the archive file (use `--verbosity DEBUG` to see traceback): '
            f'{error.__class__.__name__}:{error}'
        )

    echo.echo_success(f'migrated the archive to version {version!r}')


class ExtrasImportCode(Enum):
    """Exit codes for the verdi command line."""
    # pylint: disable=invalid-name
    keep_existing = ('k', 'c', 'l')
    update_existing = ('k', 'c', 'u')
    mirror = ('n', 'c', 'u')
    none = ('k', 'n', 'l')


@verdi_archive.command('import')
@click.argument('archives', nargs=-1, type=PathOrUrl(exists=True, readable=True))
@click.option(
    '-w',
    '--webpages',
    type=click.STRING,
    cls=options.MultipleValueOption,
    help='Discover all URL targets pointing to files with the .aiida extension for these HTTP addresses. '
    'Automatically discovered archive URLs will be downloaded and added to ARCHIVES for importing.'
)
@click.option(
    '--import-group/--no-import-group',
    default=True,
    show_default=True,
    help='Add all imported nodes to the specified group, or an automatically created one'
)
@options.GROUP(
    type=GroupParamType(create_if_not_exist=True),
    help='Specify group to which all the import nodes will be added. If such a group does not exist, it will be'
    ' created automatically.'
)
@click.option(
    '-e',
    '--extras-mode-existing',
    type=click.Choice(EXTRAS_MODE_EXISTING),
    default='none',
    help='Specify which extras from the export archive should be imported for nodes that are already contained in the '
    'database: '
    'none: do not import any extras.'
    'keep_existing: import all extras and keep original value of existing extras. '
    'update_existing: import all extras and overwrite value of existing extras. '
    'mirror: import all extras and remove any existing extras that are not present in the archive. '
)
@click.option(
    '-n',
    '--extras-mode-new',
    type=click.Choice(EXTRAS_MODE_NEW),
    default='import',
    help='Specify whether to import extras of new nodes: '
    'import: import extras. '
    'none: do not import extras.'
)
@click.option(
    '--comment-mode',
    type=click.Choice(COMMENT_MODE),
    default='leave',
    help='Specify the way to import Comments with identical UUIDs: '
    'leave: Leave the existing Comments in the database (default).'
    'newest: Use only the newest Comments (based on mtime).'
    'overwrite: Replace existing Comments with those from the import file.'
)
@click.option(
    '--include-authinfos/--exclude-authinfos',
    default=False,
    show_default=True,
    help='Include or exclude authentication information for computer(s) in import.'
)
@click.option(
    '--migration/--no-migration',
    default=True,
    show_default=True,
    help='Force migration of archive file archives, if needed.'
)
@click.option(
    '-b', '--batch-size', default=1000, type=int, help='Stream database rows in batches, to reduce memory usage.'
)
@click.option('--test-run', is_flag=True, help='Determine entities to import, but do not actually import them.')
@decorators.with_dbenv()
@click.pass_context
def import_archive(
    ctx, archives, webpages, extras_mode_existing, extras_mode_new, comment_mode, include_authinfos, migration,
    batch_size, import_group, group, test_run
):
    """Import archived data to a profile.

    The archive can be specified by its relative or absolute file path, or its HTTP URL.
    """
    # pylint: disable=unused-argument
    from aiida.common.progress_reporter import set_progress_bar_tqdm, set_progress_reporter

    if AIIDA_LOGGER.level <= logging.REPORT:  # pylint: disable=no-member
        set_progress_bar_tqdm(leave=(AIIDA_LOGGER.level <= logging.INFO))
    else:
        set_progress_reporter(None)

    all_archives = _gather_imports(archives, webpages)

    # Preliminary sanity check
    if not all_archives:
        echo.echo_critical('no valid exported archives were found')

    # Shared import key-word arguments
    import_kwargs = {
        'import_new_extras': extras_mode_new == 'import',
        'merge_extras': ExtrasImportCode[extras_mode_existing].value,
        'merge_comments': comment_mode,
        'include_authinfos': include_authinfos,
        'batch_size': batch_size,
        'create_group': import_group,
        'group': group,
        'test_run': test_run,
    }

    for archive, web_based in all_archives:
        _import_archive_and_migrate(archive, web_based, import_kwargs, migration)


def _echo_exception(msg: str, exception, warn_only: bool = False):
    """Correctly report and exception.

    :param msg: The message prefix
    :param exception: the exception raised
    :param warn_only: If True only print a warning, otherwise calls sys.exit with a non-zero exit status

    """
    from aiida.tools.archive.imports import IMPORT_LOGGER
    message = f'{msg}: {exception.__class__.__name__}: {str(exception)}'
    if warn_only:
        echo.echo_warning(message)
    else:
        IMPORT_LOGGER.info('%s', traceback.format_exc())
        echo.echo_critical(message)


def _gather_imports(archives, webpages) -> List[Tuple[str, bool]]:
    """Gather archives to import and sort into local files and URLs.

    :returns: list of (archive path, whether it is web based)

    """
    from aiida.tools.archive.common import get_valid_import_links

    final_archives = []

    # Build list of archives to be imported
    for archive in archives:
        if archive.startswith('http://') or archive.startswith('https://'):
            final_archives.append((archive, True))
        else:
            final_archives.append((archive, False))

    # Discover and retrieve *.aiida files at URL(s)
    if webpages is not None:
        for webpage in webpages:
            try:
                echo.echo_report(f'retrieving archive URLS from {webpage}')
                urls = get_valid_import_links(webpage)
            except Exception as error:
                echo.echo_critical(
                    f'an exception occurred while trying to discover archives at URL {webpage}:\n{error}'
                )
            else:
                echo.echo_success(f'{len(urls)} archive URLs discovered and added')
                final_archives.extend([(u, True) for u in urls])

    return final_archives


def _import_archive_and_migrate(archive: str, web_based: bool, import_kwargs: dict, try_migration: bool):
    """Perform the archive import.

    :param archive: the path or URL to the archive
    :param web_based: If the archive needs to be downloaded first
    :param import_kwargs: keyword arguments to pass to the import function
    :param try_migration: whether to try a migration if the import raises `IncompatibleStorageSchema`

    """
    from aiida.common.folders import SandboxFolder
    from aiida.tools.archive.abstract import get_format
    from aiida.tools.archive.imports import import_archive as _import_archive

    archive_format = get_format()

    with SandboxFolder() as temp_folder:

        archive_path = archive

        if web_based:
            echo.echo_report(f'downloading archive: {archive}')
            try:
                with urllib.request.urlopen(archive) as response:
                    temp_folder.create_file_from_filelike(response, 'downloaded_archive.zip')
            except Exception as exception:
                _echo_exception(f'downloading archive {archive} failed', exception)

            archive_path = temp_folder.get_abs_path('downloaded_archive.zip')
            echo.echo_success('archive downloaded, proceeding with import')

        echo.echo_report(f'starting import: {archive}')
        try:
            _import_archive(archive_path, archive_format=archive_format, **import_kwargs)
        except IncompatibleStorageSchema as exception:
            if try_migration:

                echo.echo_report(f'incompatible version detected for {archive}, trying migration')
                try:
                    new_path = temp_folder.get_abs_path('migrated_archive.aiida')
                    archive_format.migrate(archive_path, new_path, archive_format.latest_version, compression=0)
                    archive_path = new_path
                except Exception as exception:
                    _echo_exception(f'an exception occurred while migrating the archive {archive}', exception)

                echo.echo_report('proceeding with import of migrated archive')
                try:
                    _import_archive(archive_path, archive_format=archive_format, **import_kwargs)
                except Exception as exception:
                    _echo_exception(
                        f'an exception occurred while trying to import the migrated archive {archive}', exception
                    )
            else:
                _echo_exception(f'an exception occurred while trying to import the archive {archive}', exception)
        except Exception as exception:
            _echo_exception(f'an exception occurred while trying to import the archive {archive}', exception)

        echo.echo_success(f'imported archive {archive}')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=too-many-locals,too-many-branches,too-many-statements
"""Create an AiiDA archive.

The archive is a subset of the provenance graph,
stored in a single file.
"""
from datetime import datetime
from pathlib import Path
import shutil
import tempfile
from typing import Callable, Dict, Iterable, List, Optional, Sequence, Set, Tuple, Union

from tabulate import tabulate

from aiida import orm
from aiida.common.exceptions import LicensingException
from aiida.common.lang import type_check
from aiida.common.links import GraphTraversalRules
from aiida.common.log import AIIDA_LOGGER
from aiida.common.progress_reporter import get_progress_reporter
from aiida.manage import get_manager
from aiida.orm.entities import EntityTypes
from aiida.orm.implementation import StorageBackend
from aiida.orm.utils.links import LinkQuadruple
from aiida.tools.graph.graph_traversers import get_nodes_export, validate_traversal_rules

from .abstract import ArchiveFormatAbstract, ArchiveWriterAbstract
from .common import batch_iter, entity_type_to_orm
from .exceptions import ArchiveExportError, ExportValidationError
from .implementations.sqlite_zip import ArchiveFormatSqlZip

__all__ = ('create_archive', 'EXPORT_LOGGER')

EXPORT_LOGGER = AIIDA_LOGGER.getChild('export')
QbType = Callable[[], orm.QueryBuilder]


def create_archive(
    entities: Optional[Iterable[Union[orm.Computer, orm.Node, orm.Group, orm.User]]],
    filename: Union[None, str, Path] = None,
    *,
    archive_format: Optional[ArchiveFormatAbstract] = None,
    overwrite: bool = False,
    include_comments: bool = True,
    include_logs: bool = True,
    include_authinfos: bool = False,
    allowed_licenses: Optional[Union[list, Callable]] = None,
    forbidden_licenses: Optional[Union[list, Callable]] = None,
    strip_checkpoints: bool = True,
    batch_size: int = 1000,
    compression: int = 6,
    test_run: bool = False,
    backend: Optional[StorageBackend] = None,
    **traversal_rules: bool
) -> Path:
    """Export AiiDA data to an archive file.

    The export follows the following logic:

    First gather all entity primary keys (per type) that needs to be exported.
    This need to proceed in the "reverse" order of relationships:

    - groups: input groups
    - group_to_nodes: from nodes in groups
    - nodes & links: from graph_traversal(input nodes & group_to_nodes)
    - computers: from input computers & computers of nodes
    - authinfos: from authinfos of computers
    - comments: from comments of nodes
    - logs: from logs of nodes
    - users: from users of nodes, groups, comments & authinfos

    Now stream the full entities (per type) to the archive writer,
    in the order of relationships:

    - users
    - computers
    - authinfos
    - groups
    - nodes
    - comments
    - logs
    - group_to_nodes
    - links

    Finally stream the repository files,
    for the exported nodes, to the archive writer.

    Note, the logging level and progress reporter should be set externally, for example::

        from aiida.common.progress_reporter import set_progress_bar_tqdm

        EXPORT_LOGGER.setLevel('DEBUG')
        set_progress_bar_tqdm(leave=True)
        create_archive(...)

    :param entities: If ``None``, import all entities,
        or a list of entity instances that can include Computers, Groups, and Nodes.

    :param filename: the filename (possibly including the absolute path)
        of the file on which to export.

    :param overwrite: if True, overwrite the output file without asking, if it exists.
        If False, raise an
        :py:class:`~aiida.tools.archive.exceptions.ArchiveExportError`
        if the output file already exists.

    :param allowed_licenses: List or function.
        If a list, then checks whether all licenses of Data nodes are in the list. If a function,
        then calls function for licenses of Data nodes expecting True if license is allowed, False
        otherwise.

    :param forbidden_licenses: List or function. If a list,
        then checks whether all licenses of Data nodes are in the list. If a function,
        then calls function for licenses of Data nodes expecting True if license is allowed, False
        otherwise.

    :param include_comments: In-/exclude export of comments for given node(s) in ``entities``.
        Default: True, *include* comments in export (as well as relevant users).

    :param include_logs: In-/exclude export of logs for given node(s) in ``entities``.
        Default: True, *include* logs in export.

    :param strip_checkpoints: Remove checkpoint keys from process node attributes.
        These contain serialized code and can cause security issues.

    :param compression: level of compression to use (integer from 0 to 9)

    :param batch_size: batch database query results in sub-collections to reduce memory usage

    :param test_run: if True, do not write to file

    :param backend: the backend to export from. If not specified, the default backend is used.

    :param traversal_rules: graph traversal rules. See :const:`aiida.common.links.GraphTraversalRules`
        what rule names are toggleable and what the defaults are.

    :raises `~aiida.tools.archive.exceptions.ArchiveExportError`:
        if there are any internal errors when exporting.
    :raises `~aiida.common.exceptions.LicensingException`:
        if any node is licensed under forbidden license.

    """
    # check the backend
    backend = backend or get_manager().get_profile_storage()
    type_check(backend, StorageBackend)
    # create a function to get a query builder instance for the backend
    querybuilder = lambda: orm.QueryBuilder(backend=backend)

    # check/set archive file path
    type_check(filename, (str, Path), allow_none=True)
    if filename is None:
        filename = Path.cwd() / 'export_data.aiida'
    filename = Path(filename)
    if not overwrite and filename.exists():
        raise ArchiveExportError(f"The output file '{filename}' already exists")
    if filename.exists() and not filename.is_file():
        raise ArchiveExportError(f"The output file '{filename}' exists as a directory")

    if compression not in range(10):
        raise ArchiveExportError('compression must be an integer between 0 and 9')

    # check file format
    archive_format = archive_format or ArchiveFormatSqlZip()
    type_check(archive_format, ArchiveFormatAbstract)

    # check traversal rules
    validate_traversal_rules(GraphTraversalRules.EXPORT, **traversal_rules)
    full_traversal_rules = {
        name: traversal_rules.get(name, rule.default) for name, rule in GraphTraversalRules.EXPORT.value.items()
    }

    initial_summary = get_init_summary(
        archive_version=archive_format.latest_version,
        outfile=filename,
        collect_all=entities is None,
        include_authinfos=include_authinfos,
        include_comments=include_comments,
        include_logs=include_logs,
        traversal_rules=full_traversal_rules,
        compression=compression
    )
    EXPORT_LOGGER.report(initial_summary)

    # Store starting UUIDs, to write to metadata
    starting_uuids: Dict[EntityTypes, Set[str]] = {
        EntityTypes.USER: set(),
        EntityTypes.COMPUTER: set(),
        EntityTypes.GROUP: set(),
        EntityTypes.NODE: set()
    }

    # Store all entity IDs to be written to the archive
    # Note, this is the order they will be written to the archive
    entity_ids: Dict[EntityTypes, Set[int]] = {
        ent: set() for ent in [
            EntityTypes.USER,
            EntityTypes.COMPUTER,
            EntityTypes.AUTHINFO,
            EntityTypes.GROUP,
            EntityTypes.NODE,
            EntityTypes.COMMENT,
            EntityTypes.LOG,
        ]
    }

    # extract ids/uuid from initial entities
    type_check(entities, Iterable, allow_none=True)
    if entities is None:
        group_nodes, link_data = _collect_all_entities(
            querybuilder, entity_ids, include_authinfos, include_comments, include_logs, batch_size
        )
    else:
        for entry in entities:
            if isinstance(entry, orm.Group):
                starting_uuids[EntityTypes.GROUP].add(entry.uuid)
                entity_ids[EntityTypes.GROUP].add(entry.pk)
            elif isinstance(entry, orm.Node):
                starting_uuids[EntityTypes.NODE].add(entry.uuid)
                entity_ids[EntityTypes.NODE].add(entry.pk)
            elif isinstance(entry, orm.Computer):
                starting_uuids[EntityTypes.COMPUTER].add(entry.uuid)
                entity_ids[EntityTypes.COMPUTER].add(entry.pk)
            elif isinstance(entry, orm.User):
                starting_uuids[EntityTypes.USER].add(entry.email)
                entity_ids[EntityTypes.USER].add(entry.pk)
            else:
                raise ArchiveExportError(
                    f'I was given {entry} ({type(entry)}),'
                    ' which is not a User, Node, Computer, or Group instance'
                )
        group_nodes, link_data = _collect_required_entities(
            querybuilder, entity_ids, traversal_rules, include_authinfos, include_comments, include_logs, backend,
            batch_size
        )

    # now all the nodes have been retrieved, perform some checks
    if entity_ids[EntityTypes.NODE]:
        EXPORT_LOGGER.report('Validating Nodes')
        _check_unsealed_nodes(querybuilder, entity_ids[EntityTypes.NODE], batch_size)
        _check_node_licenses(
            querybuilder, entity_ids[EntityTypes.NODE], allowed_licenses, forbidden_licenses, batch_size
        )

    # get a count of entities, to report
    entity_counts = {etype.value: len(ids) for etype, ids in entity_ids.items()}
    entity_counts[EntityTypes.LINK.value] = len(link_data)
    entity_counts[EntityTypes.GROUP_NODE.value] = len(group_nodes)
    count_summary = [[(name + 's'), num] for name, num in entity_counts.items() if num]

    if test_run:
        EXPORT_LOGGER.report('Test Run: Stopping before archive creation')
        keys = set(
            orm.Node.collection(backend).iter_repo_keys(
                filters={'id': {
                    'in': list(entity_ids[EntityTypes.NODE])
                }}, batch_size=batch_size
            )
        )
        count_summary.append(['Repository Files', len(keys)])
        EXPORT_LOGGER.report(f'Archive would be created with:\n{tabulate(count_summary)}')
        return filename

    EXPORT_LOGGER.report(f'Creating archive with:\n{tabulate(count_summary)}')

    # Create and open the archive for writing.
    # We create in a temp dir then move to final place at end,
    # so that the user cannot end up with a half written archive on errors
    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_filename = Path(tmpdir) / 'export.zip'
        with archive_format.open(tmp_filename, mode='x', compression=compression) as writer:
            # add metadata
            writer.update_metadata({
                'ctime': datetime.now().isoformat(),
                'creation_parameters': {
                    'entities_starting_set': None if entities is None else
                    {etype.value: list(unique) for etype, unique in starting_uuids.items() if unique},
                    'include_authinfos': include_authinfos,
                    'include_comments': include_comments,
                    'include_logs': include_logs,
                    'graph_traversal_rules': full_traversal_rules,
                }
            })
            # stream entity data to the archive
            with get_progress_reporter()(desc='Archiving database: ', total=sum(entity_counts.values())) as progress:
                for etype, ids in entity_ids.items():
                    if etype == EntityTypes.NODE and strip_checkpoints:

                        def transform(row):
                            data = row['entity']
                            if data.get('node_type', '').startswith('process.'):
                                data['attributes'].pop(orm.ProcessNode.CHECKPOINT_KEY, None)
                            return data
                    else:
                        transform = lambda row: row['entity']
                    progress.set_description_str(f'Archiving database: {etype.value}s')
                    if ids:
                        for nrows, rows in batch_iter(
                            querybuilder().append(
                                entity_type_to_orm[etype], filters={
                                    'id': {
                                        'in': ids
                                    }
                                }, tag='entity', project=['**']
                            ).iterdict(batch_size=batch_size), batch_size, transform
                        ):
                            writer.bulk_insert(etype, rows)
                            progress.update(nrows)

                # stream links
                progress.set_description_str(f'Archiving database: {EntityTypes.LINK.value}s')
                transform = lambda d: {
                    'input_id': d.source_id,
                    'output_id': d.target_id,
                    'label': d.link_label,
                    'type': d.link_type
                }
                for nrows, rows in batch_iter(link_data, batch_size, transform):
                    writer.bulk_insert(EntityTypes.LINK, rows, allow_defaults=True)
                    progress.update(nrows)
                del link_data  # release memory

                # stream group_nodes
                progress.set_description_str(f'Archiving database: {EntityTypes.GROUP_NODE.value}s')
                transform = lambda d: {'dbgroup_id': d[0], 'dbnode_id': d[1]}
                for nrows, rows in batch_iter(group_nodes, batch_size, transform):
                    writer.bulk_insert(EntityTypes.GROUP_NODE, rows, allow_defaults=True)
                    progress.update(nrows)
                del group_nodes  # release memory

            # stream node repository files to the archive
            if entity_ids[EntityTypes.NODE]:
                _stream_repo_files(archive_format.key_format, writer, entity_ids[EntityTypes.NODE], backend, batch_size)

            EXPORT_LOGGER.report('Finalizing archive creation...')

        if filename.exists():
            filename.unlink()
        shutil.move(tmp_filename, filename)  # type: ignore[arg-type]

    EXPORT_LOGGER.report('Archive created successfully')

    return filename


def _collect_all_entities(
    querybuilder: QbType, entity_ids: Dict[EntityTypes, Set[int]], include_authinfos: bool, include_comments: bool,
    include_logs: bool, batch_size: int
) -> Tuple[List[list], Set[LinkQuadruple]]:
    """Collect all entities.

    :returns: (group_id_to_node_id, link_data) and updates entity_ids
    """
    progress_str = lambda name: f'Collecting entities: {name}'
    with get_progress_reporter()(desc=progress_str(''), total=9) as progress:

        progress.set_description_str(progress_str('Nodes'))
        entity_ids[EntityTypes.NODE].update(
            querybuilder().append(orm.Node, project='id').all(batch_size=batch_size, flat=True)
        )
        progress.update()

        progress.set_description_str(progress_str('Links'))
        progress.update()
        qbuilder = querybuilder().append(orm.Node, tag='incoming', project=[
            'id'
        ]).append(orm.Node, with_incoming='incoming', project=['id'], edge_project=['type', 'label']).distinct()
        link_data = {LinkQuadruple(*row) for row in qbuilder.all(batch_size=batch_size)}

        progress.set_description_str(progress_str('Groups'))
        progress.update()
        entity_ids[EntityTypes.GROUP].update(
            querybuilder().append(orm.Group, project='id').all(batch_size=batch_size, flat=True)
        )
        progress.set_description_str(progress_str('Nodes-Groups'))
        progress.update()
        qbuilder = querybuilder().append(orm.Group, project='id',
                                         tag='group').append(orm.Node, with_group='group', project='id').distinct()
        group_nodes = qbuilder.all(batch_size=batch_size)

        progress.set_description_str(progress_str('Computers'))
        progress.update()
        entity_ids[EntityTypes.COMPUTER].update(
            querybuilder().append(orm.Computer, project='id').all(batch_size=batch_size, flat=True)
        )

        progress.set_description_str(progress_str('AuthInfos'))
        progress.update()
        if include_authinfos:
            entity_ids[EntityTypes.AUTHINFO].update(
                querybuilder().append(orm.AuthInfo, project='id').all(batch_size=batch_size, flat=True)
            )

        progress.set_description_str(progress_str('Logs'))
        progress.update()
        if include_logs:
            entity_ids[EntityTypes.LOG].update(
                querybuilder().append(orm.Log, project='id').all(batch_size=batch_size, flat=True)
            )

        progress.set_description_str(progress_str('Comments'))
        progress.update()
        if include_comments:
            entity_ids[EntityTypes.COMMENT].update(
                querybuilder().append(orm.Comment, project='id').all(batch_size=batch_size, flat=True)
            )

        progress.set_description_str(progress_str('Users'))
        progress.update()
        entity_ids[EntityTypes.USER].update(
            querybuilder().append(orm.User, project='id').all(batch_size=batch_size, flat=True)
        )

    return group_nodes, link_data


def _collect_required_entities(
    querybuilder: QbType, entity_ids: Dict[EntityTypes, Set[int]], traversal_rules: Dict[str, bool],
    include_authinfos: bool, include_comments: bool, include_logs: bool, backend: StorageBackend, batch_size: int
) -> Tuple[List[list], Set[LinkQuadruple]]:
    """Collect required entities, given a set of starting entities and provenance graph traversal rules.

    :returns: (group_id_to_node_id, link_data) and updates entity_ids
    """
    progress_str = lambda name: f'Collecting entities: {name}'
    with get_progress_reporter()(desc=progress_str(''), total=7) as progress:

        # get all nodes from groups
        progress.set_description_str(progress_str('Nodes (groups)'))
        group_nodes = []
        if entity_ids[EntityTypes.GROUP]:
            qbuilder = querybuilder()
            qbuilder.append(
                orm.Group, filters={'id': {
                    'in': list(entity_ids[EntityTypes.GROUP])
                }}, project='id', tag='group'
            )
            qbuilder.append(orm.Node, with_group='group', project='id')
            qbuilder.distinct()
            group_nodes = qbuilder.all(batch_size=batch_size)
            entity_ids[EntityTypes.NODE].update(nid for _, nid in group_nodes)

        # get full set of nodes & links, following traversal rules
        progress.set_description_str(progress_str('Nodes (traversal)'))
        progress.update()
        traverse_output = get_nodes_export(
            starting_pks=entity_ids[EntityTypes.NODE], get_links=True, backend=backend, **traversal_rules
        )
        entity_ids[EntityTypes.NODE].update(traverse_output.pop('nodes'))
        link_data = traverse_output.pop('links') or set()  # possible memory hog?

        progress.set_description_str(progress_str('Computers'))
        progress.update()

        # get full set of computers
        if entity_ids[EntityTypes.NODE]:
            entity_ids[EntityTypes.COMPUTER].update(
                pk for pk, in querybuilder().append(
                    orm.Node, filters={
                        'id': {
                            'in': list(entity_ids[EntityTypes.NODE])
                        }
                    }, tag='node'
                ).append(orm.Computer, with_node='node', project='id').distinct().iterall(batch_size=batch_size)
            )

        # get full set of authinfos
        progress.set_description_str(progress_str('AuthInfos'))
        progress.update()
        if include_authinfos and entity_ids[EntityTypes.COMPUTER]:
            entity_ids[EntityTypes.AUTHINFO].update(
                pk for pk, in querybuilder().append(
                    orm.Computer, filters={
                        'id': {
                            'in': list(entity_ids[EntityTypes.COMPUTER])
                        }
                    }, tag='comp'
                ).append(orm.AuthInfo, with_computer='comp', project='id').distinct().iterall(batch_size=batch_size)
            )

        # get full set of logs
        progress.set_description_str(progress_str('Logs'))
        progress.update()
        if include_logs and entity_ids[EntityTypes.NODE]:
            entity_ids[EntityTypes.LOG].update(
                pk for pk, in querybuilder().append(
                    orm.Node, filters={
                        'id': {
                            'in': list(entity_ids[EntityTypes.NODE])
                        }
                    }, tag='node'
                ).append(orm.Log, with_node='node', project='id').distinct().iterall(batch_size=batch_size)
            )

        # get full set of comments
        progress.set_description_str(progress_str('Comments'))
        progress.update()
        if include_comments and entity_ids[EntityTypes.NODE]:
            entity_ids[EntityTypes.COMMENT].update(
                pk for pk, in querybuilder().append(
                    orm.Node, filters={
                        'id': {
                            'in': list(entity_ids[EntityTypes.NODE])
                        }
                    }, tag='node'
                ).append(orm.Comment, with_node='node', project='id').distinct().iterall(batch_size=batch_size)
            )

        # get full set of users
        progress.set_description_str(progress_str('Users'))
        progress.update()
        if entity_ids[EntityTypes.NODE]:
            entity_ids[EntityTypes.USER].update(
                pk for pk, in querybuilder().append(
                    orm.Node, filters={
                        'id': {
                            'in': list(entity_ids[EntityTypes.NODE])
                        }
                    }, tag='node'
                ).append(orm.User, with_node='node', project='id').distinct().iterall(batch_size=batch_size)
            )
        if entity_ids[EntityTypes.GROUP]:
            entity_ids[EntityTypes.USER].update(
                pk for pk, in querybuilder().append(
                    orm.Group, filters={
                        'id': {
                            'in': list(entity_ids[EntityTypes.GROUP])
                        }
                    }, tag='group'
                ).append(orm.User, with_group='group', project='id').distinct().iterall(batch_size=batch_size)
            )
        if entity_ids[EntityTypes.COMMENT]:
            entity_ids[EntityTypes.USER].update(
                pk for pk, in querybuilder().append(
                    orm.Comment, filters={
                        'id': {
                            'in': list(entity_ids[EntityTypes.COMMENT])
                        }
                    }, tag='comment'
                ).append(orm.User, with_comment='comment', project='id').distinct().iterall(batch_size=batch_size)
            )
        if entity_ids[EntityTypes.AUTHINFO]:
            entity_ids[EntityTypes.USER].update(
                pk for pk, in querybuilder().append(
                    orm.AuthInfo, filters={
                        'id': {
                            'in': list(entity_ids[EntityTypes.AUTHINFO])
                        }
                    }, tag='auth'
                ).append(orm.User, with_authinfo='auth', project='id').distinct().iterall(batch_size=batch_size)
            )

        progress.update()

    return group_nodes, link_data


def _stream_repo_files(
    key_format: str, writer: ArchiveWriterAbstract, node_ids: Set[int], backend: StorageBackend, batch_size: int
) -> None:
    """Collect all repository object keys from the nodes, then stream the files to the archive."""
    keys = set(
        orm.Node.collection(backend).iter_repo_keys(filters={'id': {
            'in': list(node_ids)
        }}, batch_size=batch_size)
    )

    repository = backend.get_repository()
    if not repository.key_format == key_format:
        # Here we would have to go back and replace all the keys in the `BackendNode.repository_metadata`s
        raise NotImplementedError(
            f'Backend repository key format incompatible: {repository.key_format!r} != {key_format!r}'
        )
    with get_progress_reporter()(desc='Archiving files: ', total=len(keys)) as progress:
        for key, stream in repository.iter_object_streams(keys):
            # to-do should we use assume the key here is correct, or always re-compute and check?
            writer.put_object(stream, key=key)
            progress.update()


def _check_unsealed_nodes(querybuilder: QbType, node_ids: Set[int], batch_size: int) -> None:
    """Check no process nodes are unsealed, i.e. all processes have completed."""
    qbuilder = querybuilder().append(
        orm.ProcessNode,
        filters={
            'id': {
                'in': list(node_ids)
            },
            'attributes.sealed': {
                '!in': [True]  # better operator?
            }
        },
        project='id'
    ).distinct()
    unsealed_node_pks = qbuilder.all(batch_size=batch_size, flat=True)
    if unsealed_node_pks:
        raise ExportValidationError(
            'All ProcessNodes must be sealed before they can be exported. '
            f"Node(s) with PK(s): {', '.join(str(pk) for pk in unsealed_node_pks)} is/are not sealed."
        )


def _check_node_licenses(
    querybuilder: QbType, node_ids: Set[int], allowed_licenses: Union[None, Sequence[str], Callable],
    forbidden_licenses: Union[None, Sequence[str], Callable], batch_size: int
) -> None:
    """Check the nodes to be archived for disallowed licences."""
    if allowed_licenses is None and forbidden_licenses is None:
        return None

    # set allowed function
    if allowed_licenses is None:
        check_allowed = lambda l: True
    elif callable(allowed_licenses):

        def _check_allowed(name):
            try:
                return allowed_licenses(name)  # type: ignore
            except Exception as exc:
                raise LicensingException('allowed_licenses function error') from exc

        check_allowed = _check_allowed
    elif isinstance(allowed_licenses, Sequence):
        check_allowed = lambda l: l in allowed_licenses  # type: ignore
    else:
        raise TypeError('allowed_licenses not a list or function')

    # set forbidden function
    if forbidden_licenses is None:
        check_forbidden = lambda l: False
    elif callable(forbidden_licenses):

        def _check_forbidden(name):
            try:
                return forbidden_licenses(name)  # type: ignore
            except Exception as exc:
                raise LicensingException('forbidden_licenses function error') from exc

        check_forbidden = _check_forbidden
    elif isinstance(forbidden_licenses, Sequence):
        check_forbidden = lambda l: l in forbidden_licenses  # type: ignore
    else:
        raise TypeError('forbidden_licenses not a list or function')

    # create query
    qbuilder = querybuilder().append(
        orm.Node,
        project=['id', 'attributes.source.license'],
        filters={'id': {
            'in': list(node_ids)
        }},
    )

    for node_id, name in qbuilder.iterall(batch_size=batch_size):
        if name is None:
            continue
        if not check_allowed(name):
            raise LicensingException(
                f"Node {node_id} is licensed under '{name}' license, which is not in the list of allowed licenses"
            )
        if check_forbidden(name):
            raise LicensingException(
                f"Node {node_id} is licensed under '{name}' license, which is in the list of forbidden licenses"
            )


def get_init_summary(
    *, archive_version: str, outfile: Path, collect_all: bool, include_authinfos: bool, include_comments: bool,
    include_logs: bool, traversal_rules: dict, compression: int
) -> str:
    """Get summary for archive initialisation"""
    parameters = [['Path', str(outfile)], ['Version', archive_version], ['Compression', compression]]

    result = f"\n{tabulate(parameters, headers=['Archive Parameters', ''])}"

    inclusions = [['Computers/Nodes/Groups/Users', 'All' if collect_all else 'Selected'],
                  ['Computer Authinfos', include_authinfos], ['Node Comments', include_comments],
                  ['Node Logs', include_logs]]
    result += f"\n\n{tabulate(inclusions, headers=['Inclusion rules', ''])}"

    if not collect_all:
        rules_table = [[f"Follow links {' '.join(name.split('_'))}s", value] for name, value in traversal_rules.items()]
        result += f"\n\n{tabulate(rules_table, headers=['Traversal rules', ''])}"

    return result + '\n'



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Test the AiiDA iPython magics."""
from IPython.testing.globalipapp import get_ipython

from aiida.tools.ipython.ipython_magics import register_ipython_extension


def test_ipython_magics():
    """Test that the %aiida magic can be loaded and adds the QueryBuilder and Node variables."""
    ipy = get_ipython()
    register_ipython_extension(ipy)

    cell = """
%aiida
qb=QueryBuilder()
qb.append(Node)
qb.all()
Dict().store()
"""
    result = ipy.run_cell(cell)

    assert result.success



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`Data` sub class to represent a folder on a file system."""

from .data import Data

__all__ = ('FolderData',)


class FolderData(Data):
    """`Data` sub class to represent a folder on a file system."""

    def __init__(self, **kwargs):
        """Construct a new `FolderData` to which any files and folders can be added.

        Use the `tree` keyword to simply wrap a directory:

            folder = FolderData(tree='/absolute/path/to/directory')

        Alternatively, one can construct the node first and then use the various repository methods to add objects:

            folder = FolderData()
            folder.put_object_from_tree('/absolute/path/to/directory')
            folder.put_object_from_filepath('/absolute/path/to/file.txt')
            folder.put_object_from_filelike(filelike_object)

        :param tree: absolute path to a folder to wrap
        :type tree: str
        """
        tree = kwargs.pop('tree', None)
        super().__init__(**kwargs)
        if tree:
            self.base.repository.put_object_from_tree(tree)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Define the current configuration version and migrations."""
from typing import Any, Dict, Iterable, Optional, Protocol, Type

from aiida.common import exceptions
from aiida.common.log import AIIDA_LOGGER

__all__ = (
    'CURRENT_CONFIG_VERSION', 'OLDEST_COMPATIBLE_CONFIG_VERSION', 'get_current_version', 'check_and_migrate_config',
    'config_needs_migrating', 'upgrade_config', 'downgrade_config', 'MIGRATIONS'
)

ConfigType = Dict[str, Any]

# The expected version of the configuration file and the oldest backwards compatible configuration version.
# If the configuration file format is changed, the current version number should be upped and a migration added.
# When the configuration file format is changed in a backwards-incompatible way, the oldest compatible version should
# be set to the new current version.

CURRENT_CONFIG_VERSION = 8
OLDEST_COMPATIBLE_CONFIG_VERSION = 8

CONFIG_LOGGER = AIIDA_LOGGER.getChild('config')


class SingleMigration(Protocol):
    """A single migration of the configuration."""

    down_revision: int
    """The initial configuration version."""

    down_compatible: int
    """The initial oldest backwards compatible configuration version"""

    up_revision: int
    """The final configuration version."""

    up_compatible: int
    """The final oldest backwards compatible configuration version"""

    def upgrade(self, config: ConfigType) -> None:
        """Migrate the configuration in-place."""

    def downgrade(self, config: ConfigType) -> None:
        """Downgrade the configuration in-place."""


class Initial(SingleMigration):
    """Base migration (no-op)."""
    down_revision = 0
    down_compatible = 0
    up_revision = 1
    up_compatible = 0

    def upgrade(self, config: ConfigType) -> None:
        pass

    def downgrade(self, config: ConfigType) -> None:
        pass


class AddProfileUuid(SingleMigration):
    """Add the required values for a new default profile.

        * PROFILE_UUID

    The profile uuid will be used as a general purpose identifier for the profile, in
    for example the RabbitMQ message queues and exchanges.
    """
    down_revision = 1
    down_compatible = 0
    up_revision = 2
    up_compatible = 0

    def upgrade(self, config: ConfigType) -> None:
        from uuid import uuid4  # we require this import here, to patch it in the tests
        for profile in config.get('profiles', {}).values():
            profile.setdefault('PROFILE_UUID', uuid4().hex)

    def downgrade(self, config: ConfigType) -> None:
        # leave the uuid present, so we could migrate back up
        pass


class SimplifyDefaultProfiles(SingleMigration):
    """Replace process specific default profiles with single default profile key.

    The concept of a different 'process' for a profile has been removed and as such the default profiles key in the
    configuration no longer needs a value per process ('verdi', 'daemon'). We remove the dictionary 'default_profiles'
    and replace it with a simple value 'default_profile'.
    """
    down_revision = 2
    down_compatible = 0
    up_revision = 3
    up_compatible = 3

    def upgrade(self, config: ConfigType) -> None:
        from aiida.manage.configuration import get_profile

        global_profile = get_profile()
        default_profiles = config.pop('default_profiles', None)

        if default_profiles and 'daemon' in default_profiles:
            config['default_profile'] = default_profiles['daemon']
        elif default_profiles and 'verdi' in default_profiles:
            config['default_profile'] = default_profiles['verdi']
        elif global_profile is not None:
            config['default_profile'] = global_profile.name

    def downgrade(self, config: ConfigType) -> None:
        if 'default_profile' in config:
            default = config.pop('default_profile')
            config['default_profiles'] = {'daemon': default, 'verdi': default}


class AddMessageBroker(SingleMigration):
    """Add the configuration for the message broker, which was not configurable up to now."""
    down_revision = 3
    down_compatible = 3
    up_revision = 4
    up_compatible = 3

    def upgrade(self, config: ConfigType) -> None:
        from aiida.manage.external.rmq import BROKER_DEFAULTS
        defaults = [
            ('broker_protocol', BROKER_DEFAULTS.protocol),
            ('broker_username', BROKER_DEFAULTS.username),
            ('broker_password', BROKER_DEFAULTS.password),
            ('broker_host', BROKER_DEFAULTS.host),
            ('broker_port', BROKER_DEFAULTS.port),
            ('broker_virtual_host', BROKER_DEFAULTS.virtual_host),
        ]

        for profile in config.get('profiles', {}).values():
            for key, default in defaults:
                if key not in profile:
                    profile[key] = default

    def downgrade(self, config: ConfigType) -> None:
        pass


class SimplifyOptions(SingleMigration):
    """Remove unnecessary difference between file/internal representation of options"""
    down_revision = 4
    down_compatible = 3
    up_revision = 5
    up_compatible = 5

    conversions = (
        ('runner_poll_interval', 'runner.poll.interval'),
        ('daemon_default_workers', 'daemon.default_workers'),
        ('daemon_timeout', 'daemon.timeout'),
        ('daemon_worker_process_slots', 'daemon.worker_process_slots'),
        ('db_batch_size', 'db.batch_size'),
        ('verdi_shell_auto_import', 'verdi.shell.auto_import'),
        ('logging_aiida_log_level', 'logging.aiida_loglevel'),
        ('logging_db_log_level', 'logging.db_loglevel'),
        ('logging_plumpy_log_level', 'logging.plumpy_loglevel'),
        ('logging_kiwipy_log_level', 'logging.kiwipy_loglevel'),
        ('logging_paramiko_log_level', 'logging.paramiko_loglevel'),
        ('logging_alembic_log_level', 'logging.alembic_loglevel'),
        ('logging_sqlalchemy_loglevel', 'logging.sqlalchemy_loglevel'),
        ('logging_circus_log_level', 'logging.circus_loglevel'),
        ('user_email', 'autofill.user.email'),
        ('user_first_name', 'autofill.user.first_name'),
        ('user_last_name', 'autofill.user.last_name'),
        ('user_institution', 'autofill.user.institution'),
        ('show_deprecations', 'warnings.showdeprecations'),
        ('task_retry_initial_interval', 'transport.task_retry_initial_interval'),
        ('task_maximum_attempts', 'transport.task_maximum_attempts'),
    )

    def upgrade(self, config: ConfigType) -> None:
        for current, new in self.conversions:
            # replace in profile options
            for profile in config.get('profiles', {}).values():
                if current in profile.get('options', {}):
                    profile['options'][new] = profile['options'].pop(current)
            # replace in global options
            if current in config.get('options', {}):
                config['options'][new] = config['options'].pop(current)

    def downgrade(self, config: ConfigType) -> None:
        for current, new in self.conversions:
            # replace in profile options
            for profile in config.get('profiles', {}).values():
                if new in profile.get('options', {}):
                    profile['options'][current] = profile['options'].pop(new)
            # replace in global options
            if new in config.get('options', {}):
                config['options'][current] = config['options'].pop(new)


class AbstractStorageAndProcess(SingleMigration):
    """Move the storage config under a top-level "storage" key and rabbitmq config under "processing".

    This allows for different storage backends to have different configuration.
    """
    down_revision = 5
    down_compatible = 5
    up_revision = 6
    up_compatible = 6

    storage_conversions = (
        ('AIIDADB_ENGINE', 'database_engine'),
        ('AIIDADB_HOST', 'database_hostname'),
        ('AIIDADB_PORT', 'database_port'),
        ('AIIDADB_USER', 'database_username'),
        ('AIIDADB_PASS', 'database_password'),
        ('AIIDADB_NAME', 'database_name'),
        ('AIIDADB_REPOSITORY_URI', 'repository_uri'),
    )
    process_keys = (
        'broker_protocol',
        'broker_username',
        'broker_password',
        'broker_host',
        'broker_port',
        'broker_virtual_host',
        'broker_parameters',
    )

    def upgrade(self, config: ConfigType) -> None:
        for profile_name, profile in config.get('profiles', {}).items():
            profile.setdefault('storage', {})
            if 'AIIDADB_BACKEND' not in profile:
                CONFIG_LOGGER.warning(f'profile {profile_name!r} had no expected "AIIDADB_BACKEND" key')
            profile['storage']['backend'] = profile.pop('AIIDADB_BACKEND', None)
            profile['storage'].setdefault('config', {})
            for old, new in self.storage_conversions:
                if old in profile:
                    profile['storage']['config'][new] = profile.pop(old)
                else:
                    CONFIG_LOGGER.warning(f'profile {profile_name!r} had no expected {old!r} key')
            profile.setdefault('process_control', {})
            profile['process_control']['backend'] = 'rabbitmq'
            profile['process_control'].setdefault('config', {})
            for key in self.process_keys:
                if key in profile:
                    profile['process_control']['config'][key] = profile.pop(key)
                elif key not in ('broker_parameters', 'broker_virtual_host'):
                    CONFIG_LOGGER.warning(f'profile {profile_name!r} had no expected {old!r} key')

    def downgrade(self, config: ConfigType) -> None:
        for profile_name, profile in config.get('profiles', {}).items():
            profile['AIIDADB_BACKEND'] = profile.get('storage', {}).get('backend', None)
            if profile['AIIDADB_BACKEND'] is None:
                CONFIG_LOGGER.warning(f'profile {profile_name!r} had no expected "storage.backend" key')
            for old, new in self.storage_conversions:
                if new in profile.get('storage', {}).get('config', {}):
                    profile[old] = profile['storage']['config'].pop(new)
            profile.pop('storage', None)
            for key in self.process_keys:
                if key in profile.get('process_control', {}).get('config', {}):
                    profile[key] = profile['process_control']['config'].pop(key)
            profile.pop('process_control', None)


class MergeStorageBackendTypes(SingleMigration):
    """`django` and `sqlalchemy` are now merged into `psql_dos`.

    The legacy name is stored under the `_v6_backend` key, to allow for downgrades.
    """
    down_revision = 6
    down_compatible = 6
    up_revision = 7
    up_compatible = 7

    def upgrade(self, config: ConfigType) -> None:
        for profile_name, profile in config.get('profiles', {}).items():
            if 'storage' in profile:
                storage = profile['storage']
                if 'backend' in storage:
                    if storage['backend'] in ('django', 'sqlalchemy'):
                        profile['storage']['_v6_backend'] = storage['backend']
                        storage['backend'] = 'psql_dos'
                    else:
                        CONFIG_LOGGER.warning(
                            f'profile {profile_name!r} had unknown storage backend {storage["backend"]!r}'
                        )

    def downgrade(self, config: ConfigType) -> None:
        for profile_name, profile in config.get('profiles', {}).items():
            if '_v6_backend' in profile.get('storage', {}):
                profile.setdefault('storage', {})['backend'] = profile.pop('_v6_backend')
            else:
                CONFIG_LOGGER.warning(f'profile {profile_name!r} had no expected "storage._v6_backend" key')


class AddTestProfileKey(SingleMigration):
    """Add the ``test_profile`` key."""
    down_revision = 7
    down_compatible = 7
    up_revision = 8
    up_compatible = 8

    def upgrade(self, config: ConfigType) -> None:
        for profile_name, profile in config.get('profiles', {}).items():
            profile['test_profile'] = profile_name.startswith('test_')

    def downgrade(self, config: ConfigType) -> None:
        profiles = config.get('profiles', {})
        profile_names = list(profiles.keys())

        # Iterate over the fixed list of the profile names, since we are mutating the profiles dictionary.
        for profile_name in profile_names:

            profile = profiles.pop(profile_name)
            profile_name_new = None
            test_profile = profile.pop('test_profile', False)  # If absent, assume it is not a test profile

            if test_profile and not profile_name.startswith('test_'):
                profile_name_new = f'test_{profile_name}'
                CONFIG_LOGGER.warning(
                    f'profile `{profile_name}` is a test profile but does not start with the required `test_` prefix.'
                )

            if not test_profile and profile_name.startswith('test_'):
                profile_name_new = profile_name[5:]
                CONFIG_LOGGER.warning(
                    f'profile `{profile_name}` is not a test profile but starts with the `test_` prefix.'
                )

            if profile_name_new is not None:

                if profile_name_new in profile_names:
                    raise exceptions.ConfigurationError(
                        f'cannot change `{profile_name}` to `{profile_name_new}` because it already exists.'
                    )

                CONFIG_LOGGER.warning(f'changing profile name from `{profile_name}` to `{profile_name_new}`.')
                profile_name = profile_name_new

            profiles[profile_name] = profile


MIGRATIONS = (
    Initial,
    AddProfileUuid,
    SimplifyDefaultProfiles,
    AddMessageBroker,
    SimplifyOptions,
    AbstractStorageAndProcess,
    MergeStorageBackendTypes,
    AddTestProfileKey,
)


def get_current_version(config):
    """Return the current version of the config.

    :return: current config version or 0 if not defined
    """
    return config.get('CONFIG_VERSION', {}).get('CURRENT', 0)


def get_oldest_compatible_version(config):
    """Return the current oldest compatible version of the config.

    :return: current oldest compatible config version or 0 if not defined
    """
    return config.get('CONFIG_VERSION', {}).get('OLDEST_COMPATIBLE', 0)


def upgrade_config(
    config: ConfigType,
    target: int = CURRENT_CONFIG_VERSION,
    migrations: Iterable[Type[SingleMigration]] = MIGRATIONS
) -> ConfigType:
    """Run the registered configuration migrations up to the target version.

    :param config: the configuration dictionary
    :return: the migrated configuration dictionary
    """
    current = get_current_version(config)
    used = []
    while current < target:
        current = get_current_version(config)
        try:
            migrator = next(m for m in migrations if m.down_revision == current)
        except StopIteration:
            raise exceptions.ConfigurationError(f'No migration found to upgrade version {current}')
        if migrator in used:
            raise exceptions.ConfigurationError(f'Circular migration detected, upgrading to {target}')
        used.append(migrator)
        migrator().upgrade(config)
        current = migrator.up_revision
        config.setdefault('CONFIG_VERSION', {})['CURRENT'] = current
        config['CONFIG_VERSION']['OLDEST_COMPATIBLE'] = migrator.up_compatible
    if current != target:
        raise exceptions.ConfigurationError(f'Could not upgrade to version {target}, current version is {current}')
    return config


def downgrade_config(
    config: ConfigType, target: int, migrations: Iterable[Type[SingleMigration]] = MIGRATIONS
) -> ConfigType:
    """Run the registered configuration migrations down to the target version.

    :param config: the configuration dictionary
    :return: the migrated configuration dictionary
    """
    current = get_current_version(config)
    used = []
    while current > target:
        current = get_current_version(config)
        try:
            migrator = next(m for m in migrations if m.up_revision == current)
        except StopIteration:
            raise exceptions.ConfigurationError(f'No migration found to downgrade version {current}')
        if migrator in used:
            raise exceptions.ConfigurationError(f'Circular migration detected, downgrading to {target}')
        used.append(migrator)
        migrator().downgrade(config)
        config.setdefault('CONFIG_VERSION', {})['CURRENT'] = current = migrator.down_revision
        config['CONFIG_VERSION']['OLDEST_COMPATIBLE'] = migrator.down_compatible
    if current != target:
        raise exceptions.ConfigurationError(f'Could not downgrade to version {target}, current version is {current}')
    return config


def check_and_migrate_config(config, filepath: Optional[str] = None):
    """Checks if the config needs to be migrated, and performs the migration if needed.

    :param config: the configuration dictionary
    :param filepath: the path to the configuration file (optional, for error reporting)
    :return: the migrated configuration dictionary
    """
    if config_needs_migrating(config, filepath):
        config = upgrade_config(config)

    return config


def config_needs_migrating(config, filepath: Optional[str] = None):
    """Checks if the config needs to be migrated.

    If the oldest compatible version of the configuration is higher than the current configuration version defined
    in the code, the config cannot be used and so the function will raise.

    :param filepath: the path to the configuration file (optional, for error reporting)
    :return: True if the configuration has an older version and needs to be migrated, False otherwise
    :raises aiida.common.ConfigurationVersionError: if the config's oldest compatible version is higher than the current
    """
    current_version = get_current_version(config)
    oldest_compatible_version = get_oldest_compatible_version(config)

    if oldest_compatible_version > CURRENT_CONFIG_VERSION:
        filepath = filepath if filepath else ''
        raise exceptions.ConfigurationVersionError(
            f'The configuration file has version {current_version} '
            f'which is not compatible with the current version {CURRENT_CONFIG_VERSION}: {filepath}\n'
            'Use a newer version of AiiDA to downgrade this configuration.'
        )

    return CURRENT_CONFIG_VERSION > current_version



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member
"""Move `Code` `hidden` attribute from `db_dbextra` to `db_dbattribute`.

Revision ID: django_0008
Revises: django_0007

"""
from alembic import op

revision = 'django_0008'
down_revision = 'django_0007'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    # The 'hidden' property of AbstractCode has been changed from an attribute to an extra
    # Therefore we find all nodes of type Code and if they have an attribute with the key 'hidden'
    # we move that value to the extra table
    #
    # First we copy the 'hidden' attributes from code.Code. nodes to the db_extra table
    op.execute(
        """
        INSERT INTO db_dbextra (key, datatype, tval, fval, ival, bval, dval, dbnode_id) (
            SELECT db_dbattribute.key, db_dbattribute.datatype, db_dbattribute.tval, db_dbattribute.fval,
            db_dbattribute.ival, db_dbattribute.bval, db_dbattribute.dval, db_dbattribute.dbnode_id
            FROM db_dbattribute JOIN db_dbnode ON db_dbnode.id = db_dbattribute.dbnode_id
            WHERE db_dbattribute.key = 'hidden'
                AND db_dbnode.type = 'code.Code.'
        );
    """
    )
    # Secondly, we delete the original entries from the DbAttribute table
    op.execute(
        """
        DELETE FROM db_dbattribute
        WHERE id in (
            SELECT db_dbattribute.id
            FROM db_dbattribute
            JOIN db_dbnode ON db_dbnode.id = db_dbattribute.dbnode_id
            WHERE db_dbattribute.key = 'hidden' AND db_dbnode.type = 'code.Code.'
        );
    """
    )


def downgrade():
    """Migrations for the downgrade."""
    raise NotImplementedError('Downgrade of django_0008.')
